"""
MIT License

Copyright (c) 2020 Dominik Kopczynski   -   dominik.kopczynski {at} univie.ac.at
                   Nils Hoffmann  -  nils.hoffmann {at} cebitec.uni-bielefeld.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


from pygoslin.parser.LipidBaseParserEventHandler import LipidBaseParserEventHandler
from pygoslin.domain.LipidAdduct import LipidAdduct
from pygoslin.domain.LipidLevel import LipidLevel
from pygoslin.domain.LipidClass import *
from pygoslin.domain.Adduct import Adduct

from pygoslin.domain.LipidFaBondType import LipidFaBondType
from pygoslin.domain.FattyAcid import FattyAcid
from pygoslin.domain.HeadGroup import HeadGroup
from pygoslin.domain.FunctionalGroup import *
from pygoslin.domain.Cycle import Cycle

from pygoslin.domain.LipidCompleteStructure import LipidCompleteStructure
from pygoslin.domain.LipidFullStructure import LipidFullStructure
from pygoslin.domain.LipidStructureDefined import LipidStructureDefined
from pygoslin.domain.LipidSnPosition import LipidSnPosition
from pygoslin.domain.LipidMolecularSpecies import LipidMolecularSpecies
from pygoslin.domain.LipidSpecies import LipidSpecies
from pygoslin.domain.LipidSpeciesInfo import LipidSpeciesInfo

from pygoslin.domain.LipidExceptions import *

class GoslinParserEventHandler(LipidBaseParserEventHandler):
    
    mediator_FA = {'H': 17, 'O': 18, 'E': 20, 'Do': 22, 'D': 22}
    mediator_DB = {'M': 1, 'D': 2, 'Tr': 3, 'tr': 3, 'T': 4, 'P': 5, 'H': 6}
    
    def __init__(self):
        super().__init__()
        self.reset_lipid(None)
        
        self.registered_events["lipid_pre_event"] = self.reset_lipid
        self.registered_events["lipid_post_event"] = self.build_lipid
        
        self.registered_events["hg_cl_pre_event"] = self.set_head_group_name
        self.registered_events["hg_mlcl_pre_event"] = self.set_head_group_name
        self.registered_events["hg_pl_pre_event"] = self.set_head_group_name
        self.registered_events["hg_lpl_pre_event"] = self.set_lpl_head_group_name
        self.registered_events["hg_lsl_pre_event"] = self.set_head_group_name
        self.registered_events["hg_so_lsl_pre_event"] = self.set_head_group_name
        self.registered_events["hg_dsl_pre_event"] = self.set_head_group_name
        self.registered_events["hg_mgl_pre_event"] = self.set_head_group_name
        self.registered_events["hg_dgl_pre_event"] = self.set_head_group_name
        self.registered_events["hg_sgl_pre_event"] = self.set_head_group_name
        self.registered_events["hg_tgl_pre_event"] = self.set_head_group_name
        self.registered_events["hg_dlcl_pre_event"] = self.set_head_group_name
        self.registered_events["hg_sac_di_pre_event"] = self.set_head_group_name
        self.registered_events["hg_sac_f_pre_event"] = self.set_head_group_name
        self.registered_events["hg_tpl_pre_event"] = self.set_head_group_name
        self.registered_events["st_pre_event"] = self.set_head_group_name
        self.registered_events["hg_ste_pre_event"] = self.set_head_group_name
        self.registered_events["hg_stes_pre_event"] = self.set_head_group_name
        
        self.registered_events["gl_species_pre_event"] = self.set_species_level
        self.registered_events["pl_species_pre_event"] = self.set_species_level
        self.registered_events["sl_species_pre_event"] = self.set_species_level
        self.registered_events["fa2_unsorted_pre_event"] = self.set_molecular_subspecies_level
        self.registered_events["fa3_unsorted_pre_event"] = self.set_molecular_subspecies_level
        self.registered_events["fa4_unsorted_pre_event"] = self.set_molecular_subspecies_level
        self.registered_events["slbpa_pre_event"] = self.set_molecular_subspecies_level
        self.registered_events["dlcl_pre_event"] = self.set_molecular_subspecies_level
        self.registered_events["mlcl_pre_event"] = self.set_molecular_subspecies_level
        
        self.registered_events["lcb_pre_event"] = self.new_lcb
        self.registered_events["lcb_post_event"] = self.clean_lcb
        self.registered_events["fa_pre_event"] = self.new_fa
        self.registered_events["fa_post_event"] = self.append_fa
        
        self.registered_events["db_single_position_pre_event"] = self.set_isomeric_level
        self.registered_events["db_single_position_post_event"] = self.add_db_position
        self.registered_events["db_position_number_pre_event"] = self.add_db_position_number
        self.registered_events["cistrans_pre_event"] = self.add_cistrans
        
        self.registered_events["ether_pre_event"] = self.add_ether
        self.registered_events["old_hydroxyl_pre_event"] = self.add_old_hydroxyl
        self.registered_events["db_count_pre_event"] = self.add_double_bonds
        self.registered_events["carbon_pre_event"] = self.add_carbon
        self.registered_events["hydroxyl_pre_event"] = self.add_hydroxyl
        self.registered_events["tpl_post_event"] = self.set_nape
        
        
        self.registered_events["adduct_info_pre_event"] = self.new_adduct
        self.registered_events["adduct_pre_event"] = self.add_adduct
        self.registered_events["charge_pre_event"] = self.add_charge
        self.registered_events["charge_sign_pre_event"] = self.add_charge_sign
        self.registered_events["plasmalogen_pre_event"] = self.add_plasmalogen
        
        self.registered_events["mediator_pre_event"] = self.set_mediator
        self.registered_events["mediator_post_event"] = self.add_mediator
        self.registered_events["unstructured_mediator_pre_event"] = self.set_unstructured_mediator
        self.registered_events["trivial_mediator_pre_event"] = self.set_trivial_mediator
        self.registered_events["mediator_carbon_pre_event"] = self.set_mediator_carbon
        self.registered_events["mediator_db_pre_event"] = self.set_mediator_db
        self.registered_events["mediator_mono_functions_pre_event"] = self.set_mediator_function
        self.registered_events["mediator_di_functions_pre_event"] = self.set_mediator_function
        self.registered_events["mediator_tri_functions_pre_event"] = self.set_mediator_function
        self.registered_events["mediator_position_pre_event"] = self.set_mediator_function_position
        self.registered_events["mediator_functional_group_post_event"] = self.add_mediator_function
        self.registered_events["mediator_suffix_pre_event"] = self.add_mediator_suffix
        self.registered_events["mediator_tetranor_pre_event"] = self.set_mediator_tetranor
        
        
        self.registered_events["isotope_pair_pre_event"] = self.new_adduct
        self.registered_events["isotope_element_pre_event"] = self.set_heavy_d_element
        self.registered_events["isotope_number_pre_event"] = self.set_heavy_d_number
        self.registered_events["heavy_pre_event"] = self.new_adduct
        self.registered_events["adduct_heavy_element_pre_event"] = self.set_heavy_element
        self.registered_events["adduct_heavy_number_pre_event"] = self.set_heavy_number
        self.registered_events["adduct_heavy_component_post_event"] = self.add_heavy_component

        self.registered_events["prostaglandin_post_event"] = self.add_prostaglandin
        self.registered_events["prostaglandin_type_pre_event"] = self.set_prostaglandin_type
        self.registered_events["prostaglandin_number_pre_event"] = self.set_prostaglandin_number

        self.debug = ""
            


    def reset_lipid(self, node):
        self.level = LipidLevel.FULL_STRUCTURE
        self.head_group = ""
        self.lcb = None
        self.fa_list = []
        self.current_fa = None
        self.adduct = None
        self.db_position = 0
        self.db_cistrans = ""
        self.unspecified_ether = False
        self.db_numbers = -1
        self.headgroup_decorators = []
        self.plasmalogen = ""
        self.mediator_function = ""
        self.mediator_function_positions = []
        self.mediator_suffix = False
        self.use_head_group = False
        self.heavy_element = None
        self.heavy_element_number = 0
        self.trivial_mediator = False
        self.prostaglandin = {"type": "", "number": ""}
        
        
        
    def add_prostaglandin(self, node):
        if self.prostaglandin["type"] not in {"B", "D", "E", "F", "J", "K"} or self.prostaglandin["number"] not in {"1", "2", "3"}: return
    
        db = {}
        tmp_fa = self.current_fa
        if self.prostaglandin["number"] == "1": db = {13: "E"}
        elif self.prostaglandin["number"] == "2": db = {5: "Z", 13: "E"}
        elif self.prostaglandin["number"] == "3": db = {5: "Z", 13: "E", 17: "Z"}
            
        if self.prostaglandin["type"] == "B":
            f1, f2, f3 = get_functional_group("OH"), get_functional_group("oxo")
            f1.position = 15
            f2.position = 9
            fg = {"OH": [f1], "cy": [Cycle(5, 8, 12, 0, functional_groups = {"oxo": [f2]})]}
            self.current_fa = FattyAcid("FA", 20, db, functional_groups = fg)
            
        elif self.prostaglandin["type"] == "D":
            f1, f2, f3 = get_functional_group("OH"), get_functional_group("OH"), get_functional_group("oxo")
            f1.position = 15
            f2.position = 9
            f3.position = 11
            fg = {"OH": [f1], "cy": [Cycle(5, 8, 12, 0, functional_groups = {"OH": [f2], "oxo": [f3]})]}
            self.current_fa = FattyAcid("FA", 20, db, functional_groups = fg)
        
        elif self.prostaglandin["type"] == "E":
            f1, f2, f3 = get_functional_group("OH"), get_functional_group("oxo"), get_functional_group("OH")
            f1.position = 15
            f2.position = 9
            f3.position = 11
            fg = {"OH": [f1], "cy": [Cycle(5, 8, 12, 0, functional_groups = {"OH": [f3], "oxo": [f2]})]}
            self.current_fa = FattyAcid("FA", 20, db, functional_groups = fg)
            
        elif self.prostaglandin["type"] == "F":
            f1, f2, f3 = get_functional_group("OH"), get_functional_group("OH"), get_functional_group("OH")
            f1.position = 15
            f2.position = 9
            f3.position = 11
            fg = {"OH": [f1], "cy": [Cycle(5, 8, 12, 0, functional_groups = {"OH": [f2, f3]})]}
            self.current_fa = FattyAcid("FA", 20, db, functional_groups = fg)
        
        
        elif self.prostaglandin["type"] == "J":
            f1, f2 = get_functional_group("OH"), get_functional_group("oxo")
            f1.position = 15
            f2.position = 11
            fg = {"OH": [f1], "cy": [Cycle(5, 8, 12, 1, functional_groups = {"OH": [f2]})]}
            self.current_fa = FattyAcid("FA", 20, db, functional_groups = fg)
            
        elif self.prostaglandin["type"] == "K":
            f1, f2, f3 = get_functional_group("OH"), get_functional_group("oxo"), get_functional_group("oxo")
            f1.position = 15
            f2.position = 9
            f3.position = 11
            fg = {"OH": [f1], "cy": [Cycle(5, 8, 12, 0, functional_groups = {"oxo": [f2, f3]})]}
            self.current_fa = FattyAcid("FA", 20, db, functional_groups = fg)
            
        else:
            return
        
        self.clean_mediator(tmp_fa)
        self.fa_list = [self.current_fa]
        self.mediator_suffix = True
        self.prostaglandin = {"type": "", "number": ""}
        
        
        
    def set_prostaglandin_type(self, node):
        self.prostaglandin["type"] = node.get_text()
        
        
        
    def set_prostaglandin_number(self, node):
        self.prostaglandin["number"] = node.get_text()
        
        
        
    def set_mediator(self, node):
        self.head_group = "FA"
        self.current_fa = FattyAcid("FA")
        self.fa_list.append(self.current_fa)
        
        
        
    def set_unstructured_mediator(self, node):
        self.head_group = node.get_text()
        self.use_head_group = True
        self.fa_list = []
        
        
    def remove_deoxy(self, functional_groups):
        if "d" in functional_groups: del functional_groups["d"]
        for fg_name, fg_list in functional_groups.items():
            for fg in fg_list:
                self.remove_deoxy(fg.functional_groups)
        
        
    def clean_mediator(self, tmp_fa):
        if len(tmp_fa.functional_groups) > 0:
            for fg_name, fg_list in tmp_fa.functional_groups.items():
                for fg in fg_list:
                    if fg.position >= -1:
                        # erase prevously added functional group, if findable
                        def recursive_deletion(fg_name, fg, functional_groups):
                            for fg_name_curr, fg_list_curr in functional_groups.items():
                                del_fg = [i for i, fg_curr in enumerate(fg_list_curr) if fg_curr.position == fg.position]
                                if len(del_fg) > 0:
                                    for i in del_fg[::-1]: fg_list_curr.pop(i)
                                    if fg_name not in functional_groups: functional_groups[fg_name] = []
                                    functional_groups[fg_name].append(fg)
                                    return True
                                
                                for fg_curr in fg_list_curr:
                                    if recursive_deletion(fg_name, fg, fg_curr.functional_groups): return True
                            return False
                                
                        if not recursive_deletion(fg_name, fg, self.current_fa.functional_groups):
                            if fg_name not in self.current_fa.functional_groups: self.current_fa.functional_groups[fg_name] = []
                            self.current_fa.functional_groups[fg_name].append(fg)
                                    
                    else:
                        if fg_name not in self.current_fa.functional_groups: self.current_fa.functional_groups[fg_name] = []
                        self.current_fa.functional_groups[fg_name].append(fg)
        
        self.remove_deoxy(self.current_fa.functional_groups)
        
        
    def set_trivial_mediator(self, node):
        self.head_group = "FA"
        mediator_name = node.get_text()
            
        tmp_fa = self.current_fa
        self.current_fa = self.resolve_fa_synonym(mediator_name)
        self.clean_mediator(tmp_fa)
        self.fa_list = [self.current_fa]
        self.mediator_suffix = True
        
        
            
        
        
        
    def set_mediator_tetranor(self, node):
        self.current_fa.num_carbon -= 4
    
    
    
    def set_mediator_carbon(self, node):
        self.trivial_mediator = True
        self.current_fa.num_carbon += GoslinParserEventHandler.mediator_FA[node.get_text()]
        
        

    def set_mediator_db(self, node):
        self.current_fa.double_bonds = GoslinParserEventHandler.mediator_DB[node.get_text()]
        
        
        
    def set_mediator_function(self, node):
        self.mediator_function = node.get_text()
        
        
        
    def set_mediator_function_position(self, node):
        self.mediator_function_positions.append(int(node.get_text()))
        
        
        
    def add_mediator_function(self, node):
        functional_group, fg = None, ""
        if self.mediator_function in {"H", "hydroxy"}:
            functional_group, fg = get_functional_group("OH"), "OH"
            if len(self.mediator_function_positions) > 0: functional_group.position = self.mediator_function_positions[0]
            
        elif self.mediator_function.lower() in {"d", "deoxy"}:
            functional_group, fg = get_functional_group("d"), "d"
            if len(self.mediator_function_positions) == 0: functional = None
            else: functional_group.position = self.mediator_function_positions[0]
            
        elif self.mediator_function.lower() in {"oxo", "k", "keto"}:
            functional_group, fg = get_functional_group("oxo"), "oxo"
            if len(self.mediator_function_positions) > 0: functional_group.position = self.mediator_function_positions[0]
            
        elif self.mediator_function.lower() == "hp":
            functional_group, fg = get_functional_group("OOH"), "OOH"
            if len(self.mediator_function_positions) > 0: functional_group.position = self.mediator_function_positions[0]
            
        elif self.mediator_function == "NO2":
            functional_group, fg = get_functional_group("NO2"), "NO2"
            if len(self.mediator_function_positions) > 0: functional_group.position = self.mediator_function_positions[0]
            
        elif self.mediator_function in {"E", "Ep", "EP"}:
            functional_group, fg = get_functional_group("Ep"), "Ep"
            if len(self.mediator_function_positions) > 0: functional_group.position = self.mediator_function_positions[0]
            
        elif self.mediator_function.lower() in {"dh", "dih", "dihydro", 'dh'}:
            functional_group, fg = get_functional_group("OH"), "OH"
            if len(self.mediator_function_positions) >= 2:
                functional_group.position = self.mediator_function_positions[0]
                functional_group2 = get_functional_group("OH")
                functional_group2.position = self.mediator_function_positions[1]
                if "OH" not in self.current_fa.functional_groups: self.current_fa.functional_groups["OH"] = []
                self.current_fa.functional_groups["OH"].append(functional_group2)
            else:
                functional_group.count = 2
            
        elif self.mediator_function.lower() in {'trih', 'trihydroxy'}:
            functional_group, fg = get_functional_group("OH"), "OH"
            if len(self.mediator_function_positions) >= 3:
                functional_group.position = self.mediator_function_positions[0]
                functional_group2 = get_functional_group("OH")
                functional_group2.position = self.mediator_function_positions[1]
                functional_group3 = get_functional_group("OH")
                functional_group3.position = self.mediator_function_positions[2]
                if "OH" not in self.current_fa.functional_groups: self.current_fa.functional_groups["OH"] = []
                self.current_fa.functional_groups["OH"].append(functional_group2)
                self.current_fa.functional_groups["OH"].append(functional_group3)
            else:
                functional_group.count = 3
                
        if functional_group != None:
            if fg not in self.current_fa.functional_groups: self.current_fa.functional_groups[fg] = []
            self.current_fa.functional_groups[fg].append(functional_group)
        
        self.mediator_function_positions = []
        self.mediator_function = ""
        
    def add_mediator_suffix(self, node):
        self.mediator_suffix = True
        
        
        
    def set_heavy_d_element(self, node):
        self.adduct.heavy_elements[Element.H2] = 1
        
        
        
    def set_heavy_d_number(self, node):
        self.adduct.heavy_elements[Element.H2] = int(node.get_text())
        
        
        
    def add_mediator(self, node):
        if not self.mediator_suffix:
            if type(self.current_fa.double_bonds) == int: self.current_fa.double_bonds -= 1
        
        
        
    def add_plasmalogen(self, node):
        plasmalogen = node.get_text().upper()
        if plasmalogen in {"O", "P"}:
            self.plasmalogen = plasmalogen
        
        
    def set_lpl_head_group_name(self, node):
        self.set_lipid_level(LipidLevel.MOLECULAR_SPECIES)
        self.head_group = node.get_text()
        
        

    def set_head_group_name(self, node):
        self.head_group = node.get_text()
        
        
        
    def set_unspecified_ether(self, node):
        self.unspecified_ether = True
        

        
    def set_nape(self, node):
        if self.head_group == "NAPE":
            self.head_group = "PE-N"
            hgd = HeadgroupDecorator("decorator_acyl", suffix = True)
            self.headgroup_decorators.append(hgd)
            hgd.functional_groups["decorator_acyl"] = [self.fa_list[-1]]
            self.fa_list.pop()
        
    
    
    def set_species_level(self, node):
        self.set_lipid_level(LipidLevel.SPECIES)
        
        
        
    def set_isomeric_level(self, node):
        self.db_position = 0
        self.db_cistrans = ""
        
        
        
    def set_structural_subspecies_level(self, node):
        self.set_lipid_level(LipidLevel.STRUCTURE_DEFINED)
        
        

    def add_db_position(self, node):
        if self.current_fa != None:
            if type(self.current_fa.double_bonds) == int:
                self.db_numbers = self.current_fa.double_bonds
                self.current_fa.double_bonds = {}
            self.current_fa.double_bonds[self.db_position] = self.db_cistrans
            if self.db_cistrans not in {"E", "Z"}: self.set_lipid_level(LipidLevel.STRUCTURE_DEFINED)
        


    def add_db_position_number(self, node):
        self.db_position = int(node.get_text())
        
        

    def add_cistrans(self, node):
        self.db_cistrans = node.get_text()
        
        
        
    def set_molecular_subspecies_level(self, node):
        self.set_lipid_level(LipidLevel.MOLECULAR_SPECIES)
        
        
        
    def new_fa(self, node):
        self.db_numbers = -1
        
        if self.unspecified_ether:
            lipid_FA_bond_type = LipidFaBondType.ETHER_UNSPECIFIED
            self.unspecified_ether = False
        else:
            lipid_FA_bond_type = LipidFaBondType.ESTER
        self.current_fa = FattyAcid("FA", lipid_FA_bond_type = lipid_FA_bond_type)

        
            
    def append_fa(self, node):
        if type(self.current_fa.double_bonds) != int:
            if self.db_numbers > -1 and self.db_numbers != len(self.current_fa.double_bonds):
                raise LipidException("Double bond count does not match with number of double bond positions")
        elif self.current_fa.double_bonds > 0:
                self.set_lipid_level(LipidLevel.SN_POSITION)
            
        if self.current_fa.lipid_FA_bond_type == LipidFaBondType.ETHER_UNSPECIFIED:
            raise LipidException("Lipid with unspecified ether bond cannot be treated properly.")
        
        if self.level in {LipidLevel.SN_POSITION, LipidLevel.STRUCTURE_DEFINED, LipidLevel.FULL_STRUCTURE, LipidLevel.COMPLETE_STRUCTURE}:
            self.current_fa.position = len(self.fa_list) + 1
        
        self.remove_deoxy(self.current_fa.functional_groups)    
        self.fa_list.append(self.current_fa)
        
        oh_cnt = 1
        if self.head_group in {'Sa', 'So', 'S1P', 'Sa1P'}:
            #self.set_lipid_level(LipidLevel.STRUCTURE_DEFINED)
            functional_group = get_functional_group("OH").copy()
            if self.head_group in {'Sa', 'So'}:
                self.fa_list[0].lipid_FA_bond_type = LipidFaBondType.LCB_EXCEPTION
                functional_group.count = 2
            else:
                self.fa_list[0].lipid_FA_bond_type = LipidFaBondType.LCB_REGULAR
                functional_group.count = 1
            if "OH" not in self.current_fa.functional_groups: self.current_fa.functional_groups["OH"] = []
            self.current_fa.functional_groups["OH"].append(functional_group)
        
        self.current_fa = None
        
        
        
    def new_lcb(self, node):
        self.lcb = FattyAcid("LCB")
        self.current_fa = self.lcb
        self.set_structural_subspecies_level(node)
        self.lcb.set_type(LipidFaBondType.LCB_REGULAR)
        self.set_lipid_level(LipidLevel.STRUCTURE_DEFINED)
            
            
            
    def clean_lcb(self, node):
        if type(self.current_fa.double_bonds) != int:
            if self.db_numbers > -1 and self.db_numbers != len(self.current_fa.double_bonds):
                raise LipidException("Double bond count does not match with number of double bond positions")
                
        elif self.current_fa.double_bonds > 0:
                self.set_lipid_level(LipidLevel.SN_POSITION)
        self.current_fa = None
        
        
        
    def build_lipid(self, node):
        if self.lcb != None:
            self.fa_list = [self.lcb] + self.fa_list
            
        if self.plasmalogen != "" and self.lcb == None and len(self.fa_list) > 0:
            self.fa_list[0].lipid_FA_bond_type = LipidFaBondType.ETHER_PLASMANYL if self.plasmalogen == "O" else LipidFaBondType.ETHER_PLASMENYL
        
        
        headgroup = self.prepare_headgroup_and_checks()
        
        if self.trivial_mediator and node.get_text() in trivial_mediators:
            if type(self.fa_list[0].double_bonds) != dict: self.fa_list[0].double_bonds = {}
            for p in trivial_mediators[node.get_text()]:
                if len(p) == 0: continue
                if p[-1] in {"E", "Z"}:
                    self.fa_list[0].double_bonds[int(p[:-1])] = p[-1]
                else:
                    self.fa_list[0].double_bonds[int(p)] = ""
            self.level = LipidLevel.FULL_STRUCTURE
            
        lipid = LipidAdduct()
        lipid.adduct = self.adduct
        lipid.lipid = self.assemble_lipid(headgroup)
        
        self.content = lipid
        
        
        
    def add_ether(self, node):
        ether = node.get_text()
        if ether == "a": self.current_fa.lipid_FA_bond_type = LipidFaBondType.ETHER_PLASMANYL
        elif ether == "p":
            self.current_fa.lipid_FA_bond_type = LipidFaBondType.ETHER_PLASMENYL
            if type(self.current_fa.double_bonds) == int:
                self.current_fa.double_bonds = max(0, self.current_fa.double_bonds - 1)
        self.plasmalogen = ""
        
        
        
    def add_old_hydroxyl(self, node):
        old_hydroxyl = node.get_text()
        
        num_h = 0
        if old_hydroxyl == "d": num_h = 2
        if old_hydroxyl == "t": num_h = 3
        
        if self.sp_regular_lcb(): num_h -= 1
        
        functional_group = get_functional_group("OH").copy()
        functional_group.count = num_h
        if "OH" not in self.current_fa.functional_groups: self.current_fa.functional_groups["OH"] = []
        self.current_fa.functional_groups["OH"].append(functional_group)
        
        
        
    def add_double_bonds(self, node):
        self.current_fa.double_bonds = int(node.get_text())
        
        
        
    def add_carbon(self, node):
        self.current_fa.num_carbon = int(node.get_text())
        
        
        
    def add_hydroxyl(self, node):
        num_h = int(node.get_text())
        if self.sp_regular_lcb(): num_h -= 1
        if num_h <= 0: return
        
        functional_group = get_functional_group("OH").copy()
        self.set_lipid_level(LipidLevel.STRUCTURE_DEFINED)
        functional_group.count = num_h
        if "OH" not in self.current_fa.functional_groups: self.current_fa.functional_groups["OH"] = []
        self.current_fa.functional_groups["OH"].append(functional_group)
        
        
        
    def new_adduct(self, node):
        if self.adduct == None: self.adduct = Adduct("", "")
        
        
        
    def add_adduct(self, node):
        self.adduct.adduct_string = node.get_text()
        
        
        
    def add_charge(self, node):
        self.adduct.charge = int (node.get_text())
        
        
        
    def add_charge_sign(self, node):
        sign = node.get_text()
        if sign == "+": self.adduct.set_charge_sign(1)
        if sign == "-": self.adduct.set_charge_sign(-1)
        if self.adduct.charge == 0: self.adduct.charge = 1
        
        
        
    def set_heavy_element(self, node):
        self.heavy_element = heavy_element_table[node.get_text()]
        self.heavy_element_number = 1
        
        
        
    def set_heavy_number(self, node):
        self.heavy_element_number = int(node.get_text())
        
        
        
    def add_heavy_component(self, node):
        self.adduct.heavy_elements[self.heavy_element] += self.heavy_element_number
        
        
trivial_mediators = {}
tm_dir_name = path.dirname(pygoslin.__file__)
tm_file_name = path.join(tm_dir_name, "data", "goslin", "trivial_mediators.csv")
with open(tm_file_name, mode = "rt", encoding= "utf-8") as tm_infile:
    for line in tm_infile:
        tokens = line.strip().split("\t")
        if len(tokens) != 2: continue
        trivial_mediators[tokens[0]] = [p for p in tokens[1].split(",")]



"""
MIT License

Copyright (c) 2020 Dominik Kopczynski   -   dominik.kopczynski {at} univie.ac.at
                   Nils Hoffmann  -  nils.hoffmann {at} cebitec.uni-bielefeld.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


from enum import Enum
from os import path

import pygoslin
from pygoslin.domain.LipidExceptions import LipidParsingException, LipidException, ParserException
from itertools import combinations as iter_combinations
from pygoslin.parser.SumFormulaParserEventHandler import SumFormulaParserEventHandler


pyx_support = True
try:
    import pyximport
    from pygoslin.parser.ParserCore import parser_core
except:
    pyx_support = False


class Context(Enum):
    NoContext = 1
    InLineComment = 2
    InLongComment = 3
    InQuote = 4
    
class MatchWords(Enum):
    NoMatch = 1
    LineCommentStart = 2
    LineCommentEnd = 3
    LongCommentStart = 4
    LongCommentEnd = 5
    Quote = 6
    
    
        

class TreeNode:

    def __init__(self, _rule, _fire_event):
        self.rule_index = _rule
        self.left = None
        self.right = None
        self.terminal = 0
        self.fire_event = _fire_event
    
    def get_text(self):
        if self.terminal == 0:
            left_str = self.left.get_text()
            right_str = self.right.get_text() if self.right != None else ""
            return "%s%s" % (left_str if left_str != Parser.EOF_SIGN else "", right_str if right_str != Parser.EOF_SIGN else "")
        return self.terminal
        
        
        
        
     
compute_rule_key = lambda rule_index_1, rule_index_2: (rule_index_1 << Parser.SHIFT) | rule_index_2
        
        
        
class Parser:
    
    SHIFT = 32
    MASK = (1 << SHIFT) - 1
    RULE_ASSIGNMENT = ':'
    RULE_SEPARATOR = '|'
    RULE_TERMINAL = ';'
    EOF_RULE_NAME = "EOF"
    EOF_SIGN = chr(1)
    EOF_RULE = 1
    START_RULE = 2
    DEFAULT_QUOTE = "'"
    
    
    
    
    def get_next_free_rule_index(self):
        if self.next_free_rule_index <= Parser.MASK:
            n = self.next_free_rule_index
            self.next_free_rule_index += 1
            return n
        raise ParserException("Error: grammar is too big.")
    
    
    
    
    
    def __init__(self, _parserEventHandler, grammar_filename = None, _quote = DEFAULT_QUOTE):
        self.next_free_rule_index = Parser.START_RULE
        self.TtoNT = {}
        self.NTtoNT = {}
        self.NTtoRule = {}
        self.quote = _quote
        self.parser_event_handler = _parserEventHandler
        self.word_in_grammar = False
        self.grammar_name = ""
        self.used_eof = False
        self.substitution = {}
        self.OTtoNT = {}
        self.right_pair = []
        self.left_pair = []
        self.error_message = ""
        
        if grammar_filename == None: return
        
        if path.exists(grammar_filename):
            # interpret the rules and create the structure for parsing
            rules = Parser.extract_text_based_rules(grammar_filename, self.quote)
            self.grammar_name = Parser.split_string(rules[0], ' ', self.quote)[1]
            del rules[0]
            ruleToNT = {Parser.EOF_RULE_NAME: Parser.EOF_RULE}
            self.TtoNT[Parser.EOF_SIGN] = set([Parser.EOF_RULE])
            for rule_line in rules:
                tokens_level_1 = []
                for t in Parser.split_string(rule_line, Parser.RULE_ASSIGNMENT, self.quote): tokens_level_1.append(t.strip(' '))
                if len(tokens_level_1) != 2: raise ParserException("Error: corrupted token in grammar rule: '%s'" % rule_line);
                
                if len(Parser.split_string(tokens_level_1[0], ' ', self.quote)) > 1:
                    raise ParserException("Error: several rule names on left hand side in grammar rule: '%s'" % rule_line);
                

                rule = tokens_level_1[0]
                
                if rule ==  Parser.EOF_RULE_NAME:
                    raise ParserException("Error: rule name is not allowed to be called EOF")
                
                products = [p.strip(' ') for p in Parser.split_string(tokens_level_1[1], Parser.RULE_SEPARATOR, self.quote)]
                
                if rule not in ruleToNT: ruleToNT[rule] = self.get_next_free_rule_index()
                new_rule_index = ruleToNT[rule]
                
                if new_rule_index not in self.NTtoRule: self.NTtoRule[new_rule_index] = rule
                
                
                for product in products:
                    if len(product) == 0:
                        raise ParserException("Error: empty product rule in rule: \"%s\"" % rule_line)
                    non_terminals, non_terminal_rules = [], []
                    for NT in Parser.split_string(product, ' ', self.quote):
                        stripedNT = NT.strip(' ')
                        if Parser.is_terminal(stripedNT, self.quote): stripedNT = Parser.de_escape(stripedNT, self.quote)
                        non_terminals.append(stripedNT)
                        self.used_eof |= stripedNT == Parser.EOF_RULE_NAME
                        
                    
                    NTFirst = non_terminals[0]
                    if len(non_terminals) > 1 or not Parser.is_terminal(NTFirst, self.quote) or len(NTFirst) != 3:
                        for non_terminal in non_terminals:
                            
                            if Parser.is_terminal(non_terminal, self.quote):
                                non_terminal_rules.append(self.add_terminal(non_terminal))
                                
                            else:
                                if non_terminal not in ruleToNT:
                                    ruleToNT[non_terminal] = self.get_next_free_rule_index()
                                non_terminal_rules.append(ruleToNT[non_terminal])
                                
                    else:
                        c = NTFirst[1]
                        t_rule = 0
                        if c not in self.TtoNT: 
                            t_rule = self.get_next_free_rule_index()
                            self.TtoNT[c] = set([t_rule])
                        else:
                            t_rule = list(self.TtoNT[c])[0]
                            
                        if t_rule not in self.NTtoNT: self.NTtoNT[t_rule] = set()
                        self.NTtoNT[t_rule].add(new_rule_index)
                    	
                    
                    # more than two rules, insert intermediate rule indexes
                    while len(non_terminal_rules) > 2:
                        rule_index_2 = non_terminal_rules.pop()
                        rule_index_1 = non_terminal_rules.pop()
                        
                        key = compute_rule_key(rule_index_1, rule_index_2)
                        next_index = self.get_next_free_rule_index()
                        if key not in self.NTtoNT: self.NTtoNT[key] = set()
                        self.NTtoNT[key].add(next_index)
                        non_terminal_rules.append(next_index)
                    
                        
                    # two product rules
                    if len(non_terminal_rules) == 2:
                        rule_index_2 = non_terminal_rules[1]
                        rule_index_1 = non_terminal_rules[0]
                        key = compute_rule_key(rule_index_1, rule_index_2)
                        if key not in self.NTtoNT: self.NTtoNT[key] = set()
                        self.NTtoNT[key].add(new_rule_index)
                    
                    # only one product rule
                    elif len(non_terminal_rules) == 1:
                        rule_index_1 = non_terminal_rules[0]
                        if rule_index_1 == new_rule_index:
                            raise ParserException("Error: corrupted token in grammar: rule '%s' is not allowed to refer soleley to itself." % rule)
                        
                        if rule_index_1 not in self.NTtoNT: self.NTtoNT[rule_index_1] = set()
                        self.NTtoNT[rule_index_1].add(new_rule_index)
            
            # adding all rule names into the event handler
            for rule_name in ruleToNT:
                self.parser_event_handler.rule_names.add(rule_name)
                
            self.parser_event_handler.parser = self
            self.parser_event_handler.sanity_check()
            
        else:
            raise ParserException("Error: file '%s' does not exist or can not be opened." % grammar_filename)
        
        def top_nodes( rule_index):
            collection, collection_top = [rule_index], []
            i = 0
            while i < len(collection):
                current_index = collection[i]
                if current_index in self.NTtoNT:
                    for previous_index in self.NTtoNT[current_index]: collection.append(previous_index)
                else:
                    collection_top.append(current_index)
                i += 1
            return collection_top
        
        
        
        visited = set()
        for rule_index, values in self.NTtoNT.items():
            rules_to_visit = (set(values) | set([rule_index])) - visited
            for rule in rules_to_visit:
                visited.add(rule)
                
                for rule_top in set(top_nodes(rule)):
                    chains = self.collect_backwards(rule, rule_top)
                    if chains == None: continue
                    
                    for chain in chains:
                        while len(chain) > 1:
                            top = chain[0]
                            chain = chain[1:]
                            self.substitution[(rule_index, top)] = chain
                    
                    
        for k, v in self.TtoNT.items():
            self.OTtoNT[k] = list(v)[0]
            
            
        
        
        for d in [self.TtoNT, self.NTtoNT]:
            for rules in d.values():
                new_rules = set()
                for rule in rules:
                    new_rules |= set([p for p in self.collect_one_backwards(rule)])
                rules |= new_rules
                    
            
    
        self.right_pair = [set() for i in range(self.next_free_rule_index)]
        self.left_pair = [set() for i in range(self.next_free_rule_index)]
        for key in self.NTtoNT:
            if key <= self.MASK: continue
            self.right_pair[key >> self.SHIFT].add(key)
            self.left_pair[key & self.MASK].add(key)
           
           

    
    
    
    
    def extract_text_based_rules(grammar_filename, quote = DEFAULT_QUOTE):
        grammar, sb,current_position = "", [], 0
        
        with open(grammar_filename, mode = "rt", encoding= "utf-8") as infile:
            grammar = infile.read() + "\n";
        grammar_length = len(grammar)
        
        # deleting comments to prepare for splitting the grammar in rules.
        # Therefore, we have to consider three different contexts, namely
        # within a quote, within a line comment, within a long comment.
        # As long as we are in one context, key words for starting / ending
        # the other contexts have to be ignored.
        current_context = Context.NoContext
        last_escaped_backslash = -1
        for i in range (grammar_length - 1):
            match = MatchWords.NoMatch
            
            if i > 0 and grammar[i] == '\\' and grammar[i - 1] == '\\' and last_escaped_backslash != i - 1:
                last_escaped_backslash = i
                continue
            
            if grammar[i] == '/' and grammar[i + 1] == '/': match = MatchWords.LineCommentStart
            elif grammar[i] == '\n': match = MatchWords.LineCommentEnd
            elif grammar[i] == '/' and grammar[i + 1] == '*': match = MatchWords.LongCommentStart
            elif grammar[i] == '*' and grammar[i + 1] == '/': match = MatchWords.LongCommentEnd
            elif grammar[i] == quote and not (i >= 1 and grammar[i - 1] == '\\' and i - 1 != last_escaped_backslash): match = MatchWords.Quote
            
            if match != MatchWords.NoMatch:
                
                if current_context == Context.NoContext:
                    if match == MatchWords.LongCommentStart:
                        sb.append(grammar[current_position : i])
                        current_context = Context.InLongComment
                        
                    elif match == MatchWords.LineCommentStart:
                        sb.append(grammar[current_position : i])
                        current_context = Context.InLineComment
                        
                    elif match == MatchWords.Quote:
                        current_context = Context.InQuote
                    
                elif current_context == Context.InQuote:
                    if match == MatchWords.Quote:
                        current_context = Context.NoContext;
                    
                    
                elif current_context == Context.InLineComment:
                    if match == MatchWords.LineCommentEnd:
                        current_context = Context.NoContext
                        current_position = i + 1
                    
                elif current_context == Context.InLongComment:
                    if match == MatchWords.LongCommentEnd:
                        current_context = Context.NoContext
                        current_position = i + 2
                
                    
                    
        if current_context == Context.NoContext:
            sb.append(grammar[current_position : grammar_length])
            
        else:
            raise ParserException("Error: corrupted grammar '%s', ends either in comment or quote" % grammar_filename)
        
        grammar = "".join(sb).replace("\r\n", "").replace("\n", "").replace("\r", "").strip(" ");
        if grammar[-1] != Parser.RULE_TERMINAL:
            raise ParserException("Error: corrupted grammar'%s', last rule has no termininating sign, was: '%s'" % (grammar_filename, grammar[-1]))
        
        rules = Parser.split_string(grammar, Parser.RULE_TERMINAL, quote)
        
        if len(rules) < 1:
            raise ParserException("Error: corrupted grammar '%s', grammar is empty" % grammar_filename)
        
        grammar_name_rule = Parser.split_string(rules[0], ' ', quote)
        
        if grammar_name_rule[0] != "grammar":
            raise ParserException("Error: first rule must start with the keyword 'grammar'")
        
        elif len(grammar_name_rule) != 2:
            raise ParserException("Error: incorrect first rule")
        
        return rules
    
    
    
    
    def dump_child(self, class_name, parser_event_handler_name):
        with open("%s.py" % class_name, mode = "wt", encoding= "utf-8") as dumpfile:
            dumpfile.write("from pygoslin.parser.Parser import Parser\n")
            dumpfile.write("from pygoslin.parser.%s import %s\n\n" % (parser_event_handler_name, parser_event_handler_name))
            dumpfile.write("class %s(Parser):\n" % class_name)
            dumpfile.write("    def __init__(self):\n")
            dumpfile.write("        super().__init__(%s(), None, Parser.DEFAULT_QUOTE)\n" % parser_event_handler_name)
            dumpfile.write("        self.next_free_rule_index = %s\n" % str(self.next_free_rule_index))
            dumpfile.write("        self.TtoNT = %s\n" % str(self.TtoNT))
            dumpfile.write("        self.NTtoNT = %s\n" % str(self.NTtoNT))
            dumpfile.write("        self.NTtoRule = %s\n" % str(self.NTtoRule))
            dumpfile.write("        self.word_in_grammar = %s\n" % str(self.word_in_grammar))
            dumpfile.write("        self.used_eof = %s\n" % str(self.used_eof))
            dumpfile.write("        self.substitution = %s\n" % str(self.substitution))
            dumpfile.write("        self.OTtoNT = %s\n" % str(self.OTtoNT))
            dumpfile.write("        self.right_pair = %s\n" % str(self.right_pair))
            dumpfile.write("        self.left_pair = %s\n" % str(self.left_pair))
    
    
    
    def split_string(text, separator, quote = DEFAULT_QUOTE, with_empty = False):
        in_quote, tokens, sb, last_char, last_escaped_backslash = False, [], [], '\0', False
        
        for c in text:
            escaped_backslash = False
            if not in_quote:
                if c == separator:
                    if len(sb) > 0 or with_empty: tokens.append("".join(sb))
                    sb = []
                else:
                    if c == quote: in_quote = not in_quote
                    sb.append(c)
                    
            else:
                if c == '\\' and last_char == '\\' and not last_escaped_backslash:
                    escaped_backslash = True
                    
                elif c == quote and not (last_char == '\\' and not last_escaped_backslash):
                    in_quote = not in_quote
                sb.append(c)
            last_escaped_backslash = escaped_backslash
            last_char = c
                
        if len(sb) > 0 or (last_char == ',' and with_empty):
            tokens.append("".join(sb))
        if in_quote: raise ParserException("Error: corrupted token in grammar")
        
        return tokens
    
    
    
    # checking if string is terminal
    is_terminal = lambda product_token, quote: product_token[0] == quote and product_token[-1] == quote and len(product_token) > 2
    
    
    
    
    def de_escape(text, quote):
        # remove the escape chars
        sb, last_escape_char = [], False
        for c in text:
            escape_char = False
            
            if c != '\\':
                sb.append(c)
                
            else:
                if not last_escape_char: escape_char = True
                else: sb.append(c)
            
            last_escape_char = escape_char
        
        return "".join(sb)
    
    
    
    # splitting the whole terminal in a tree structure where characters of terminal are the leafs and the inner nodes are added non terminal rules
    def add_terminal(self, text):
        terminal_rules = []
        for i in range(1, len(text) - 1):
            c = text[i]
            t_rule = 0
            if c not in self.TtoNT:
                t_rule = self.get_next_free_rule_index()
                self.TtoNT[c] = set([t_rule])
            else:
                t_rule = list(self.TtoNT[c])[0]
            terminal_rules.append(t_rule)
        
        while len(terminal_rules) > 1:
            rule_index_2 = terminal_rules.pop()
            rule_index_1 = terminal_rules.pop()
            
            next_index = self.get_next_free_rule_index()
            
            key = compute_rule_key(rule_index_1, rule_index_2)
            if key not in self.NTtoNT: self.NTtoNT[key] = set()
            self.NTtoNT[key].add(next_index)
            terminal_rules.append(next_index)
        
        return terminal_rules[0]
    
    
    
    
    # expanding singleton rules, e.g. S -> A, A -> B, B -> C
    def collect_one_backwards(self, rule_index):
        collection = [rule_index]
        i = 0
        while i < len(collection):
            current_index = collection[i]
            if current_index in self.NTtoNT:
                for previous_index in self.NTtoNT[current_index]: collection.append(previous_index)
            i += 1
        return collection
    
    
    
    
    def collect_backwards(self, child_rule_index, parent_rule_index, visited = None, path = None, collection = None):
        # provides all single linkage paths from a child rule to a parent rule,
        # and yes, there can be several paths
        
        if visited == None: visited, path, collection = set(), [], []
        
        if child_rule_index not in self.NTtoNT: return collection
        visited.add(child_rule_index)
        path.append(child_rule_index)
        
        for previous_rule in self.NTtoNT[child_rule_index]:
            if previous_rule not in visited:
                
                if previous_rule == parent_rule_index:
                    collection.append(list(path + [parent_rule_index])[::-1])
                
                else:
                    collection = self.collect_backwards(previous_rule, parent_rule_index, visited, path, collection)
                      
        path.pop()
        visited.remove(child_rule_index)
        return collection
        
        
        
        
    
    def raise_events(self, node):
        if node:
            node_rule_name = self.NTtoRule[node.rule_index] if node.fire_event else ""
            if node.fire_event: self.parser_event_handler.handle_event(node_rule_name + "_pre_event", node)
            
            if node.left: # node.terminal is != None when node is leaf
                self.raise_events(node.left)
                if node.right: self.raise_events(node.right)
                
            if node.fire_event: self.parser_event_handler.handle_event(node_rule_name + "_post_event", node)
            
    
    
    
    

    # filling the syntax tree including events
    def fill_tree(self, node, parse_content):
        # checking and extending nodes for single rule chains
        
        if parse_content[1] != None:
            bottom_rule = compute_rule_key(parse_content[0], parse_content[2])
            top_rule = node.rule_index
        else:
            top_rule = parse_content[2]
            bottom_rule = self.OTtoNT[parse_content[0]]
        

            
        s_key = (bottom_rule, top_rule)
        if bottom_rule != top_rule and s_key in self.substitution:
            for rule_index in self.substitution[s_key]:
                node.left = TreeNode(rule_index, rule_index in self.NTtoRule)
                node = node.left
        
        
        if parse_content[1] != None: # None => leaf
            node.left = TreeNode(parse_content[0], parse_content[0] in self.NTtoRule)
            node.right = TreeNode(parse_content[2], parse_content[2] in self.NTtoRule)
            self.fill_tree(node.left, parse_content[1])
            self.fill_tree(node.right, parse_content[3])
            
        else:
            node.terminal = parse_content[0]
    
    
    
    
    def parse(self, text_to_parse, raise_error = True):
        self.error_message = ""
        text_to_parse = text_to_parse.strip(" ")
        old_lipid = text_to_parse
        if self.used_eof: text_to_parse += Parser.EOF_SIGN
        
        try:
            self.parse_regular(text_to_parse)
            if raise_error and not self.word_in_grammar:
                raise LipidParsingException("Lipid '%s' can not be parsed by grammar '%s'" % (old_lipid, self.grammar_name))
        except Exception as ex:
            if raise_error: raise ex
        
        return self.parser_event_handler.content
        
        
        
        
        
        
    def parse_regular(self, text_to_parse):
        self.word_in_grammar = False
        self.parser_event_handler.content = None
        
        # call parser core function written in cython for performance boost
        dp_table = parser_core(text_to_parse, self.NTtoNT, self.TtoNT, self.left_pair, self.right_pair) if pyx_support else self.parser_pure(text_to_parse)
        if dp_table == None: return
        
        
        for i in range(len(text_to_parse) - 1, 0, -1):
            if Parser.START_RULE in dp_table[i][0]:
                self.word_in_grammar = True
                parse_tree = TreeNode(Parser.START_RULE, Parser.START_RULE in self.NTtoRule)
                self.fill_tree(parse_tree, dp_table[i][0][Parser.START_RULE])
                self.raise_events(parse_tree)
                break
            if self.used_eof: break
            
        # search for longest parsed prefix
        if not self.word_in_grammar:
            for i in range(len(text_to_parse) - 1, 0, -1):
                if len(dp_table[i][0]) > 0:
                    first_rule = list(dp_table[i][0].keys())[0]
                    parse_tree = TreeNode(first_rule, first_rule in self.NTtoRule)
                    self.fill_tree(parse_tree, dp_table[i][0][first_rule])
                    self.error_message = parse_tree.get_text()
                    break
                
                
    def get_error_message(self):
        return self.error_message
            
            
            
    # re-implementation of Cocke-Younger-Kasami algorithm
    def parser_pure(self, text_to_parse):
        n = len(text_to_parse)
        rgt = self.right_pair
        lft = self.left_pair
        
        
        
        DP = [[{} for j in range(n - i)] for i in range(n)]
        DL = [[set() for j in range(n - i)] for i in range(n)]
        DR = [[set() for j in range(n - i)] for i in range(n)]
        Ks = [set([0]) for i in range(n)]
        
        t, nt, mask, shift = self.TtoNT, self.NTtoNT, (1 << 32) - 1, 32
            
        for i in range(n):
            c = text_to_parse[i]
            if c not in t: return None
            
            for rule_index in t[c]:
                DP[0][i][rule_index] = [c, None, rule_index, None]
                DL[0][i] |= lft[rule_index]
                DR[0][i] |= rgt[rule_index]
        
        
        for i in range (1, n):
            im1, DPi = i - 1, DP[i]
            for j in range(n - i):
                jp1, DPij, Ksj = j + 1, DPi[j], Ks[j]
                
                for k in Ksj:
                    jpok, im1mk = jp1 + k, im1 - k
                    if im1mk not in Ks[jpok]: continue
                    
                    for key in DR[k][j] & DL[im1mk][jpok]:
                        index_pair_1 = key >> shift
                        index_pair_2 = key & mask
                        parse_content = [index_pair_1, DP[k][j][index_pair_1], index_pair_2, DP[im1mk][jpok][index_pair_2]]
                        for rule_index in nt[key]:
                            DPij[rule_index] = parse_content
                            DL[i][j] |= lft[rule_index]
                            DR[i][j] |= rgt[rule_index]
                                    
                if DPij: Ksj.add(i)

        return DP
           
           
class SumFormulaParser(Parser):
    def __init__(self):
        self.event_handler = SumFormulaParserEventHandler()
        dir_name = path.dirname(pygoslin.__file__)
        file_name = path.join(dir_name, "data", "goslin", "SumFormula.g4")
        super().__init__(self.event_handler, file_name, Parser.DEFAULT_QUOTE)

        



"""
MIT License

Copyright (c) 2020 Dominik Kopczynski   -   dominik.kopczynski {at} univie.ac.at
                   Nils Hoffmann  -  nils.hoffmann {at} cebitec.uni-bielefeld.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


import unittest
import os


try:
    import pyximport
    pyximport.install(setup_args = {"script_args" : ["--force"]}, language_level = 3)
except:
    print("Warning: cython module is not installed, parsing performance will be lower since pure python code will be applied.")

import pygoslin
from pygoslin.parser.Parser import GoslinParser
from pygoslin.domain.LipidLevel import *

class GoslinTest(unittest.TestCase):
    PARSER_QUOTE = '\''
    
    
    def test_sphingolipids(self):
        parser = GoslinParser()
        
        l = parser.parse("Cer 18:1(8Z);2/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.STRUCTURE_DEFINED), "Cer 18:1(8);(OH)2/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.SN_POSITION), "Cer 18:1;O2/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.MOLECULAR_SPECIES), "Cer 18:1;O2/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.SPECIES), "Cer 42:1;O2")
        self.assertEqual(l.get_sum_formula(), "C42H83NO3")
        
        l = parser.parse("HexCer 18:1(5Z);2/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.STRUCTURE_DEFINED), "HexCer 18:1(5);OH/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.SN_POSITION), "HexCer 18:1;O2/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.MOLECULAR_SPECIES), "HexCer 18:1;O2/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.SPECIES), "HexCer 42:1;O2")
        self.assertEqual(l.get_sum_formula(), "C48H93NO8")
        
        l = parser.parse("LSM 17:1(4E);2")
        self.assertEqual(l.get_lipid_string(LipidLevel.STRUCTURE_DEFINED), "LSM 17:1(4);OH")
        self.assertEqual(l.get_lipid_string(LipidLevel.SN_POSITION), "LSM 17:1;O2")
        self.assertEqual(l.get_lipid_string(LipidLevel.MOLECULAR_SPECIES), "LSM 17:1;O2")
        self.assertEqual(l.get_lipid_string(LipidLevel.SPECIES), "LSM 17:1;O2")
        self.assertEqual(l.get_sum_formula(), "C22H47N2O5P")
        
        l = parser.parse("LCB 18:1(4E);2")
        self.assertEqual(l.get_lipid_string(LipidLevel.STRUCTURE_DEFINED), "SPB 18:1(4);(OH)2")
        self.assertEqual(l.get_lipid_string(LipidLevel.SN_POSITION), "SPB 18:1;O2")
        self.assertEqual(l.get_lipid_string(LipidLevel.MOLECULAR_SPECIES), "SPB 18:1;O2")
        self.assertEqual(l.get_lipid_string(LipidLevel.SPECIES), "SPB 18:1;O2")
        self.assertEqual(l.get_sum_formula(), "C18H37NO2")

        l = parser.parse("EPC 14:1(4E);2/20:1(11Z)")
        self.assertEqual(l.get_lipid_string(LipidLevel.STRUCTURE_DEFINED), "EPC 14:1(4);OH/20:1(11)")
        self.assertEqual(l.get_lipid_string(LipidLevel.SN_POSITION), "EPC 14:1;O2/20:1")
        self.assertEqual(l.get_lipid_string(LipidLevel.MOLECULAR_SPECIES), "EPC 14:1;O2/20:1")
        self.assertEqual(l.get_lipid_string(LipidLevel.SPECIES), "EPC 34:2;O2")
        self.assertEqual(l.get_sum_formula(), "C36H71N2O6P")
        
        l = parser.parse("MIPC 18:0;3/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.STRUCTURE_DEFINED), "MIPC 18:0;(OH)2/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.SN_POSITION), "MIPC 18:0;O3/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.MOLECULAR_SPECIES), "MIPC 18:0;O3/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.SPECIES), "MIPC 42:0;O3")
        self.assertEqual(l.get_sum_formula(), "C54H106NO17P")
        
        l = parser.parse("EPC 16:2(4E,6E);2/22:1(13Z);1")
        self.assertEqual(l.get_lipid_string(LipidLevel.STRUCTURE_DEFINED), "EPC 16:2(4,6);OH/22:1(13);OH")
        self.assertEqual(l.get_lipid_string(LipidLevel.SN_POSITION), "EPC 16:2;O2/22:1;O")
        self.assertEqual(l.get_lipid_string(LipidLevel.MOLECULAR_SPECIES), "EPC 16:2;O2/22:1;O")
        self.assertEqual(l.get_lipid_string(LipidLevel.SPECIES), "EPC 38:3;O3")
        self.assertEqual(l.get_sum_formula(), "C40H77N2O7P")
    
    
        l = parser.parse("13-oxoODE")
        self.assertEqual(l.get_lipid_string(LipidLevel.MOLECULAR_SPECIES), "FA 18:3;O")
        self.assertEqual(l.get_sum_formula(), "C18H30O3")
        
        
    
    
    def test_parser(self):
        lipidnames = []
        file_name = os.path.join("pygoslin", "data", "goslin", "testfiles", "goslin-test.csv")
        with open(file_name, mode = "rt", encoding= "utf-8") as infile:
            for line in infile:
                line = line.strip().strip(" ")
                if len(line) > 0: lipidnames.append(line)
        
        
        lipid_parser = GoslinParser()
        
        for i, lipid_name in enumerate(lipidnames):
            try:
                lipid = lipid_parser.parse(lipid_name)
                lipid_class = lipid.get_lipid_string(LipidLevel.CLASS)
                assert lipid_class not in {"Undefined lipid class", "Undefined", "UNDEFINED"}
                
            except Exception as e:
                print("hier: '%s' -> %i" % (lipid_name, i))
                print(e)
                assert (False)



"""
MIT License

Copyright (c) 2020 Dominik Kopczynski   -   dominik.kopczynski {at} univie.ac.at
                   Nils Hoffmann  -  nils.hoffmann {at} cebitec.uni-bielefeld.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


from pygoslin.domain.LipidLevel import LipidLevel
from pygoslin.domain.Element import *

class LipidAdduct:

    def __init__(self):
        self.lipid = None
        self.adduct = None
        self.fragment = None
        self.sum_formula = None
        
        
        
    def get_lipid_string(self, level = None):
        lipid_name = []
        
        if self.lipid != None:
            if level == None: level = self.lipid.info.level
            lipid_name.append(self.lipid.get_lipid_string(level))
        else: return ""
        
        if level == None or level not in {LipidLevel.CLASS, LipidLevel.CATEGORY}:
            if self.adduct != None: lipid_name.append(self.adduct.get_lipid_string())
        
        return "".join(lipid_name)
    
    
    def get_extended_class(self):
        return self.lipid.get_extended_class() if self.lipid != None else ""
        
        
        
    def get_lipid_fragment_string(self, level = None):
        lipid_name = []
        
        if self.lipid != None: lipid_name.append(self.lipid.get_lipid_string(level))
        else: return ""
        
        if level == None or level not in {LipidLevel.CLASS, LipidLevel.CATEGORY}:
            if self.adduct != None: lipid_name.append(self.adduct.get_lipid_string())
        
            if self.fragment != None:
                lipid_name.append(" - ")
                lipid_name.append(self.fragment.get_lipid_string())
        
        return "".join(lipid_name)
    
    
    
    def get_elements(self):
        elements = {e: 0 for e in Element}
        
        if self.lipid != None:
            lipid_elements = self.lipid.get_elements()
            for e in Element: elements[e] += lipid_elements[e]
                
        if self.adduct != None:
            adduct_elements = self.adduct.get_elements()
            for e in Element:
                elements[e] += adduct_elements[e]
        return elements
    
    
    
    def get_mass(self):
        elements = self.get_elements()
        charge = charge = self.adduct.get_charge() if self.adduct != None else 0
        mass = sum([element_masses[e] * elements[e] for e in Element])
        if charge != 0: mass = (mass - charge * electron_rest_mass) / abs(charge)
            
        return mass
    
    
    
    def compute_mass(elements):
        mass = sum([element_masses[e] * elements[e] for e in Element])
        return mass
    
    
    def sort_fatty_acyl_chains(self):
        self.lipid.sort_fatty_acyl_chains()
    
    
    def get_sum_formula(self):
        return compute_sum_formula(self.get_elements())



"""
MIT License

Copyright (c) 2020 Dominik Kopczynski   -   dominik.kopczynski {at} univie.ac.at
                   Nils Hoffmann  -  nils.hoffmann {at} cebitec.uni-bielefeld.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


from pygoslin.parser.LipidBaseParserEventHandler import LipidBaseParserEventHandler
from pygoslin.domain.LipidAdduct import LipidAdduct
from pygoslin.domain.LipidLevel import LipidLevel
from pygoslin.domain.Adduct import Adduct
from pygoslin.domain.LipidFaBondType import LipidFaBondType
from pygoslin.domain.FattyAcid import FattyAcid
from pygoslin.domain.LipidCompleteStructure import LipidCompleteStructure
from pygoslin.domain.LipidFullStructure import LipidFullStructure
from pygoslin.domain.LipidStructureDefined import LipidStructureDefined
from pygoslin.domain.LipidSnPosition import LipidSnPosition
from pygoslin.domain.LipidMolecularSpecies import LipidMolecularSpecies
from pygoslin.domain.LipidSpecies import LipidSpecies
from pygoslin.domain.LipidSpeciesInfo import LipidSpeciesInfo
from pygoslin.domain.LipidExceptions import *
from pygoslin.domain.HeadGroup import HeadGroup
from pygoslin.domain.FunctionalGroup import *
from pygoslin.domain.LipidClass import *
from pygoslin.domain.Cycle import *

class HmdbParserEventHandler(LipidBaseParserEventHandler):
    
    def __init__(self):
        super().__init__()
        
        self.registered_events["lipid_pre_event"] = self.reset_lipid
        self.registered_events["lipid_post_event"] = self.build_lipid
        self.registered_events["fa_hg_pre_event"] = self.set_head_group_name
        self.registered_events["gl_hg_pre_event"] = self.set_head_group_name
        self.registered_events["gl_mono_hg_pre_event"] = self.set_head_group_name
        self.registered_events["gl_molecular_hg_pre_event"] = self.set_head_group_name
        self.registered_events["mediator_pre_event"] = self.mediator_event
        
        ## set adduct events
        self.registered_events["adduct_info_pre_event"] = self.new_adduct
        self.registered_events["adduct_pre_event"] = self.add_adduct
        self.registered_events["charge_pre_event"] = self.add_charge
        self.registered_events["charge_sign_pre_event"] = self.add_charge_sign
        
        self.registered_events["pl_hg_pre_event"] = self.set_head_group_name
        self.registered_events["pl_three_hg_pre_event"] = self.set_head_group_name
        self.registered_events["pl_four_hg_pre_event"] = self.set_head_group_name
        self.registered_events["sl_hg_pre_event"] = self.set_head_group_name
        self.registered_events["st_species_hg_pre_event"] = self.set_head_group_name
        self.registered_events["st_sub1_hg_pre_event"] = self.set_head_group_name
        self.registered_events["st_sub2_hg_pre_event"] = self.set_head_group_name
        self.registered_events["ganglioside_names_pre_event"] = self.set_head_group_name
        self.registered_events["fa_species_pre_event"] = self.set_species_level
        self.registered_events["gl_molecular_pre_event"] = self.set_molecular_level
        self.registered_events["unsorted_fa_separator_pre_event"] = self.set_molecular_level
        self.registered_events["fa2_unsorted_pre_event"] = self.set_molecular_level
        self.registered_events["fa3_unsorted_pre_event"] = self.set_molecular_level
        self.registered_events["fa4_unsorted_pre_event"] = self.set_molecular_level
        
        self.registered_events["db_single_position_pre_event"] = self.set_isomeric_level
        self.registered_events["db_single_position_post_event"] = self.add_db_position
        self.registered_events["db_position_number_pre_event"] = self.add_db_position_number
        self.registered_events["cistrans_pre_event"] = self.add_cistrans
        
        self.registered_events["lcb_pre_event"] = self.new_lcb
        self.registered_events["lcb_post_event"] = self.clean_lcb
        self.registered_events["fa_pre_event"] = self.new_fa
        self.registered_events["fa_post_event"] = self.append_fa
        self.registered_events["fa_lcb_suffix_type_pre_event"] = self.add_one_hydroxyl
        
        self.registered_events["ether_pre_event"] = self.add_ether
        self.registered_events["hydroxyl_pre_event"] = self.add_hydroxyl
        self.registered_events["db_count_pre_event"] = self.add_double_bonds
        self.registered_events["carbon_pre_event"] = self.add_carbon
        
        self.registered_events["interlink_fa_pre_event"] = self.interlink_fa
        self.registered_events["lipid_suffix_pre_event"] = self.lipid_suffix
        self.registered_events["methyl_pre_event"] = self.add_methyl
        
        self.registered_events["furan_fa_pre_event"] = self.furan_fa
        self.registered_events["furan_fa_post_event"] = self.furan_fa_post
        self.registered_events["furan_fa_mono_pre_event"] = self.furan_fa_mono
        self.registered_events["furan_fa_di_pre_event"] = self.furan_fa_di
        self.registered_events["furan_first_number_pre_event"] = self.furan_fa_first_number
        self.registered_events["furan_second_number_pre_event"] = self.furan_fa_second_number
        
        self.registered_events["fa_lcb_suffix_types_pre_event"] = self.register_suffix_type
        self.registered_events["fa_lcb_suffix_position_pre_event"] = self.register_suffix_pos
        self.registered_events["fa_synonym_pre_event"] = self.register_fa_synonym
        
    def reset_lipid(self, node):
        self.level = LipidLevel.FULL_STRUCTURE
        self.adduct = None
        self.head_group = ""
        self.lcb = None
        self.fa_list = []
        self.current_fa = None
        self.db_position = 0
        self.db_cistrans = ""
        self.use_head_group = False
        self.headgroup_decorators = []
        self.db_number = -1
        self.furan = {}
        self.func_type = ""
        self.func_position = 0
        self.update_functional_groups = []
        
        
    def register_suffix_type(self, node):
        self.func_type = node.get_text()
        if self.func_type not in {"me", "OH", "O"}:
            raise LipidException("Unknown functional abbreviation: %s" % self.func_type)
        if self.func_type == "me": self.func_type = "Me"
        elif self.func_type == "O": self.func_type = "oxo"
    
    
    
    def register_suffix_pos(self, node):
        pos = int(node.get_text())
        functional_group = get_functional_group(self.func_type).copy()
        functional_group.position = pos
        if self.func_type not in self.current_fa.functional_groups: self.current_fa.functional_groups[self.func_type] = []
        self.current_fa.functional_groups[self.func_type].append(functional_group)
            
            
    
    def register_fa_synonym(self, node):
        self.current_fa = self.resolve_fa_synonym(node.get_text())
        
            

    def add_db_position(self, node):
        if self.current_fa != None:
            if type(self.current_fa.double_bonds) == int:
                self.db_numbers = self.current_fa.double_bonds
                self.current_fa.double_bonds = {}
            self.current_fa.double_bonds[self.db_position] = self.db_cistrans
            if self.db_cistrans not in {"E", "Z"}: self.set_lipid_level(LipidLevel.STRUCTURE_DEFINED)
        

    def add_db_position_number(self, node):
        self.db_position = int(node.get_text())
        

    def add_cistrans(self, node):
        self.db_cistrans = node.get_text()
        

    def set_head_group_name(self, node):
        self.head_group = node.get_text()
        
        
    def set_isomeric_level(self, node):
        self.db_position = 0
        self.db_cistrans = ""
        
    
    def set_species_level(self, node):
        self.set_lipid_level(LipidLevel.SPECIES)
        
        
    def set_molecular_level(self, node):
        self.set_lipid_level(LipidLevel.MOLECULAR_SPECIES)
        
        
    def mediator_event(self, node):
        self.use_head_group = True
        self.head_group = node.get_text()
        
        
    def new_fa(self, node):
        self.db_numbers = -1
        self.current_fa = FattyAcid("FA")
        
        
    def new_lcb(self, node):
        self.lcb = FattyAcid("LCB")
        self.lcb.set_type(LipidFaBondType.LCB_REGULAR)
        self.set_structural_subspecies_level(node)
        self.current_fa = self.lcb
            
            
    def clean_lcb(self, node):
        if type(self.current_fa.double_bonds) != int:
            if self.db_numbers > -1 and self.db_numbers != len(self.current_fa.double_bonds):
                raise LipidException("Double bond count does not match with number of double bond positions")
        elif self.current_fa.double_bonds > 0:
                self.set_lipid_level(LipidLevel.SN_POSITION)
            
        self.current_fa = None
        
        
            
    def append_fa(self, node):
        
        if len(self.update_functional_groups) > 0:
            for fg in self.update_functional_groups:
                fg.position += self.current_fa.num_carbon
            self.update_functional_groups = []
        
        if type(self.current_fa.double_bonds) != int:
            if self.db_numbers > -1 and self.db_numbers != len(self.current_fa.double_bonds):
                raise LipidException("Double bond count does not match with number of double bond positions")
        elif self.current_fa.double_bonds > 0:
                self.set_lipid_level(LipidLevel.SN_POSITION)
            
        if self.level in {LipidLevel.SN_POSITION, LipidLevel.STRUCTURE_DEFINED, LipidLevel.FULL_STRUCTURE, LipidLevel.COMPLETE_STRUCTURE}:
            self.current_fa.position = len(self.fa_list) + 1
            
        self.fa_list.append(self.current_fa)
        self.current_fa = None
    
        
    def build_lipid(self, node):
        if self.lcb != None:
            self.fa_list = [self.lcb] + self.fa_list
        
        headgroup = self.prepare_headgroup_and_checks()
        
        lipid = LipidAdduct()
        lipid.lipid = self.assemble_lipid(headgroup)
        lipid.adduct = self.adduct
        
        self.content = lipid
        
        
    def add_ether(self, node):
        ether = node.get_text()
        if ether in {"o-", "O-"}: self.current_fa.lipid_FA_bond_type = LipidFaBondType.ETHER_PLASMANYL
        elif ether == "P-": self.current_fa.lipid_FA_bond_type = LipidFaBondType.ETHER_PLASMENYL
        else: raise UnsupportedLipidException("Fatty acyl chain of type '%s' is currently not supported" % ether)
    
    def add_methyl(self, node):
        functional_group = get_functional_group("Me").copy()
        functional_group.position = -(1 if node.get_text() == "i-" else 2)
        self.update_functional_groups.append(functional_group)
        self.current_fa.num_carbon -= 1
        if "Me" not in self.current_fa.functional_groups: self.current_fa.functional_groups["Me"] = []
        self.current_fa.functional_groups["Me"].append(functional_group)
        
        
    def set_structural_subspecies_level(self, node):
        self.set_lipid_level(LipidLevel.STRUCTURE_DEFINED)
            
        
    def add_hydroxyl(self, node):
        hydroxyl, num_h = node.get_text(), 0
        if hydroxyl == "m": num_h = 1
        if hydroxyl == "d": num_h = 2
        if hydroxyl == "t": num_h = 3
        
        if self.sp_regular_lcb(): num_h -= 1
        
        functional_group = get_functional_group("OH").copy()
        functional_group.count = num_h
        if "OH" not in self.current_fa.functional_groups: self.current_fa.functional_groups["OH"] = []
        self.current_fa.functional_groups["OH"].append(functional_group)
        
        
    def add_one_hydroxyl(self, node):
        functional_group = get_functional_group("OH").copy()
        if "OH" not in self.current_fa.functional_groups: self.current_fa.functional_groups["OH"] = []
        self.current_fa.functional_groups["OH"].append(functional_group)
        
        
    def add_double_bonds(self, node):
        self.current_fa.double_bonds = int(node.get_text())
        
        
    def add_carbon(self, node):
        self.current_fa.num_carbon += int(node.get_text())
        
        
    def furan_fa(self, node):
        self.furan = {}
        
    
    def furan_fa_post(self, node):
        l = 4 + self.furan["len_first"] + self.furan["len_second"]
        self.current_fa.num_carbon = l
        
        start = 1 + self.furan["len_first"]
        end = 4 + self.furan["len_first"]
        cyclo_db = {start: "E", 2 + start: "E"}
        cyclo_fg = {"Me": []}
        if self.furan["type"] == "m":
            fg = get_functional_group("Me")
            fg.position = 1 + start
            cyclo_fg["Me"].append(fg)
            
        elif self.furan["type"] == "d":
            fg = get_functional_group("Me")
            fg.position = 1 + start
            cyclo_fg["Me"].append(fg)
            fg = get_functional_group("Me")
            fg.position = 2 + start
            cyclo_fg["Me"].append(fg)
           
        bridge_chain = [Element.O]
        
        self.current_fa.functional_groups["cy"] = [Cycle(5, start = start, end = end, double_bonds = cyclo_db, functional_groups = cyclo_fg, bridge_chain = bridge_chain)]
        
    
    def furan_fa_mono(self, node):
        self.furan["type"] = "m"
        
    
    def furan_fa_di(self, node):
        self.furan["type"] = "d"
    
    
    def furan_fa_first_number(self, node):
        self.furan["len_first"] = int(node.get_text())
    
    
    def furan_fa_second_number(self, node):
        self.furan["len_second"] = int(node.get_text())
        
        
    def interlink_fa(self, node):
        raise UnsupportedLipidException("Interconnected fatty acyl chains are currently not supported")


    def lipid_suffix(self, node):
        pass
        #raise UnsupportedLipidException("Lipids with suffix '%s' are currently not supported" % node.get_text())
        
        
    def new_adduct(self, node):
        self.adduct = Adduct("", "")
        
        
    def add_adduct(self, node):
        self.adduct.adduct_string = node.get_text()
        
        
    def add_charge(self, node):
        self.adduct.charge = int (node.get_text())
        
        
    def add_charge_sign(self, node):
        sign = node.get_text()
        if sign == "+": self.adduct.set_charge_sign(1)
        if sign == "-": self.adduct.set_charge_sign(-1)
        if self.adduct.charge == 0: self.adduct.charge = 1



"""
MIT License

Copyright (c) 2020 Dominik Kopczynski   -   dominik.kopczynski {at} univie.ac.at
                   Nils Hoffmann  -  nils.hoffmann {at} cebitec.uni-bielefeld.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


from pygoslin.parser.BaseParserEventHandler import BaseParserEventHandler
from pygoslin.domain.LipidAdduct import LipidAdduct
from pygoslin.domain.LipidLevel import LipidLevel
from pygoslin.domain.Adduct import Adduct
from pygoslin.domain.LipidFaBondType import LipidFaBondType
from pygoslin.domain.LipidSpeciesInfo import LipidSpeciesInfo
from pygoslin.domain.LipidSpecies import LipidSpecies
from pygoslin.domain.FattyAcid import FattyAcid
from pygoslin.domain.LipidMolecularSubspecies import LipidMolecularSubspecies
from pygoslin.domain.LipidStructuralSubspecies import LipidStructuralSubspecies
from pygoslin.domain.LipidIsomericSubspecies import LipidIsomericSubspecies
from pygoslin.domain.LipidExceptions import *
from pygoslin.domain.HeadGroup import HeadGroup
from pygoslin.domain.FunctionalGroup import *
from pygoslin.domain.LipidClass import *
from pygoslin.domain.Cycle import *

last_numbers = {'un': 1, 'hen': 1, 'do': 2, 'di': 2, 'tri': 3, 'buta': 4, 'but': 4, 'tetra': 4, 'penta': 5, 'pent': 5, 'hexa': 6, 'hex': 6, 'hepta': 7, 'hept': 7, 'octa': 8, 'oct': 8, 'nona': 9, 'non': 9}

second_numbers = {'deca': 10, 'dec': 10, 'eicosa': 20, 'eicos': 20 , 'cosa': 20, 'cos': 20, 'triaconta': 30, 'triacont': 30, 'tetraconta': 40, 'tetracont': 40, 'pentaconta': 50, 'pentacont': 50, 'hexaconta': 60, 'hexacont': 60, 'heptaconta': 70, 'heptacont': 70, 'octaconta': 80, 'octacont': 80, 'nonaconta': 90, 'nonacont': 90}

special_numbers = {'meth': 1, 'etha': 2, 'eth': 2,  'propa': 3, 'isoprop': 3, 'prop': 3, 'propi': 3, 'propio': 3, 'buta': 4, 'but': 4, 'butr': 4, 'valer': 5, 'eicosa': 20, 'eicos': 20, 'icosa': 20, 'icos': 20, 'prosta': 20, 'prost': 20, 'prostan': 20}

func_groups = {'keto': 'oxo', 'ethyl': 'Et', 'hydroxy': "OH", 'phospho': 'Ph', 'oxo': 'oxo', 'bromo': 'Br', 'methyl': 'Me', 'hydroperoxy': 'OOH', 'homo': '', 'Epoxy': 'Ep', 'fluro': 'F', 'fluoro': 'F', 'chloro': 'Cl', 'methylene': 'My', 'sulfooxy': 'Su', 'amino': 'NH2', 'sulfanyl': 'SH', 'methoxy': 'OMe', 'iodo': 'I', 'cyano': 'CN', 'nitro': 'NO2', 'OH': 'OH', 'thio': 'SH', 'mercapto': 'SH', 'carboxy': "COOH", 'acetoxy': 'Ac', 'cysteinyl': 'Cys', 'phenyl': 'Phe', 's-glutathionyl': "SGlu", 's-cysteinyl': "SCys", "butylperoxy": "BOO", "dimethylarsinoyl": "MMAs", "methylsulfanyl": "SMe", "imino": "NH", 's-cysteinylglycinyl': "SCG"}

ate = {'formate': 1, 'acetate': 2, 'butyrate': 4, 'propionate': 3, 'valerate': 5, 'isobutyrate': 4}


class SystematicParserEventHandler(BaseParserEventHandler):
    
    def __init__(self):
        super().__init__()
        
        self.registered_events["lipid_pre_event"] = self.reset_lipid
        self.registered_events["lipid_post_event"] = self.build_lipid
        
        ## FA category
        self.registered_events["FA_post_event"] = self.build_FA
        
        
        ## GL category
        self.registered_events["GL_post_event"] = self.build_GL
        self.registered_events["GP_post_event"] = self.build_GL_GP_SP
        self.registered_events["generic_fa_pre_event"] = self.set_generic_fa
        self.registered_events["generic_fa2_pre_event"] = self.set_generic_fa
        self.registered_events["generic_fa3_pre_event"] = self.set_generic_fa
        self.registered_events["generic_fa_rob_post_event"] = self.add_generic_fa
        self.registered_events["generic_fa_rob_s_post_event"] = self.add_generic_fa
        self.registered_events["generic_pos_pre_event"] = self.set_generic_pos
        self.registered_events["gl_ending_pre_event"] = self.set_gl_class
        self.registered_events["xdg_xmg_ending_pre_event"] = self.set_gl_class
        self.registered_events["gp_ending_pre_event"] = self.set_gp_class
        
        
        
        
        ## fatty acyl
        self.registered_events["fa_pre_event"] = self.set_fa
        self.registered_events["fa_post_event"] = self.build_fa
        self.registered_events["fatty_acid_post_event"] = self.set_fatty_acid
        self.registered_events["acetyl_pre_event"] = self.set_acetyl
        
        self.registered_events["acid_single_type_pre_event"] = self.set_fatty_acyl_type
        self.registered_events["ol_ending_pre_event"] = self.set_fatty_acyl_type
        self.registered_events["double_bond_position_pre_event"] = self.set_double_bond_information
        self.registered_events["double_bond_position_post_event"] = self.add_double_bond_information
        self.registered_events["db_number_post_event"] = self.set_double_bond_position
        self.registered_events["cistrans_post_event"] = self.set_cistrans
        self.registered_events["acid_type_double_post_event"] = self.check_db
        
        ## lengths
        self.registered_events["functional_length_pre_event"] = self.reset_length
        self.registered_events["fatty_length_pre_event"] = self.reset_length
        self.registered_events["functional_length_post_event"] = self.set_functional_length
        self.registered_events["fatty_length_post_event"] = self.set_fatty_length
        
        ## numbers
        self.registered_events["notation_specials_pre_event"] = self.special_number
        self.registered_events["notation_last_digit_pre_event"] = self.last_number
        self.registered_events["notation_second_digit_pre_event"] = self.second_number
        
        ## functional groups
        self.registered_events["functional_group_pre_event"] = self.set_functional_group
        self.registered_events["functional_group_post_event"] = self.add_functional_group
        self.registered_events["functional_pos_pre_event"] = self.set_functional_pos
        self.registered_events["functional_position_pre_event"] = self.set_functional_position
        self.registered_events["functional_group_type_pre_event"] = self.set_functional_type
        
        ## cyclo / epoxy
        self.registered_events["cyclo_position_pre_event"] = self.set_functional_group
        self.registered_events["cyclo_position_post_event"] = self.rearrange_cycle
        self.registered_events["epoxy_pre_event"] = self.set_functional_group
        self.registered_events["epoxy_post_event"] = self.add_epoxy
        self.registered_events["cycle_pre_event"] = self.set_cycle
        self.registered_events["methylene_post_event"] = self.set_methylene
        self.registered_events["furan_pre_event"] = self.set_furan
        
        ## dioic
        self.registered_events["dioic_pre_event"] = self.set_functional_group
        self.registered_events["dioic_post_event"] = self.set_dioic
        self.registered_events["dioic_acid_pre_event"] = self.set_fatty_acyl_type
        self.registered_events["dial_post_event"] = self.set_dial
        
        
        ## prosta
        self.registered_events["prosta_pre_event"] = self.set_prosta
        self.registered_events["prosta_post_event"] = self.add_cyclo
        self.registered_events["reduction_pre_event"] = self.set_functional_group
        self.registered_events["reduction_post_event"] = self.reduction
        self.registered_events["homo_post_event"] = self.homo
        
        
        ## recursion
        self.registered_events["recursion_description_pre_event"] = self.set_recursion
        self.registered_events["recursion_description_post_event"] = self.add_recursion
        self.registered_events["recursion_pos_pre_event"] = self.set_recursion_pos
        self.registered_events["yl_ending_pre_event"] = self.set_yl_ending
        self.registered_events["acetic_acid_post_event"] = self.set_acetic_acid
        self.registered_events["acetic_recursion_pre_event"] = self.set_recursion
        self.registered_events["acetic_recursion_post_event"] = self.add_recursion
        
        self.registered_events["hydroxyl_number_pre_event"] = self.add_hydroxyl
        self.registered_events["ol_pre_event"] = self.setup_hydroxyl
        self.registered_events["ol_post_event"] = self.add_hydroxyls
        self.registered_events["ol_pos_post_event"] = self.set_yl_ending
        
        
        
        ## wax esters
        self.registered_events["wax_ester_pre_event"] = self.set_recursion
        self.registered_events["wax_ester_post_event"] = self.add_wax_ester
        self.registered_events["ate_post_event"] = self.set_ate
        self.registered_events["isoprop_post_event"] = self.set_iso
        self.registered_events["isobut_post_event"] = self.set_iso
        
        ## CoA
        self.registered_events["CoA_post_event"] = self.set_coa
        self.registered_events["methyl_pre_event"] = self.set_methyl
        
        ## CAR
        self.registered_events["CAR_pre_event"] = self.set_car
        self.registered_events["CAR_post_event"] = self.add_car
        
        ## amine
        self.registered_events["ethanolamine_post_event"] = self.add_ethanolamine
        self.registered_events["amine_n_pre_event"] = self.set_recursion
        self.registered_events["amine_n_post_event"] = self.add_amine
        self.registered_events["amine_post_event"] = self.add_amine_name
        
        ## functional group position summary
        self.registered_events["fg_pos_summary_pre_event"] = self.set_functional_group
        self.registered_events["fg_pos_summary_post_event"] = self.add_summary
        self.registered_events["func_stereo_pre_event"] = self.add_func_stereo
        
        
        
    def reset_lipid(self, node):
        self.level = LipidLevel.ISOMERIC_SUBSPECIES
        self.headgroup = ""
        self.fatty_acyl_stack = []
        self.fatty_acyl_list = []
        self.fatty_acyl_dict = {}
        self.tmp = {}
        self.fatty_acyl = None
        
        #self.debug = "full"
        
        
        
    def set_gp_class(self, node):
        gp_class = node.get_text()
        if gp_class == 'sn-glycero-3-phosphocholine': self.headgroup = "PC"
        elif gp_class == 'sn-glycero-3-phosphoethanolamine': self.headgroup = "PE"
        elif gp_class == 'sn-glycero-3-phosphoserine': self.headgroup = "PS"
        
        
        
    def set_gl_class(self, node):
        gl_class = node.get_text()
        if gl_class in {'glycerol', 'rac-glycerol', 'sn-glycerol'}: self.tmp["gl_class"] = ""
        elif gl_class == '3-o-β-d-galactosyl-sn-glycerol': self.tmp["gl_class"] = "MG"
        elif gl_class in {'3-o-(6\'-o-α-d-galactosyl-β-d-galactosyl)-sn-glycerol', '3-o-(α-d-galactosyl1-6)-β-d-galactosyl-sn-glycerol'}: self.tmp["gl_class"] = "DG"
        elif gl_class == '3-(6\'-sulfo-α-d-quinovosyl)-sn-glycerol': self.tmp["gl_class"] = "SQ"
        
        
    def set_fa(self, node):
        self.fatty_acyl_stack.append(FattyAcid("FA"))
        self.tmp["fa1"] = {"fa_type": ""}
        


    def build_FA(self, node):
        self.headgroup = self.tmp["fa1"]["fa_type"]
        self.fatty_acyl_list = self.fatty_acyl_stack
        
        
    
    def build_fa(self, node):        
        if "cyclo_yl" in self.tmp:
            self.tmp["fg_pos"] = [[1, ""], [self.tmp["cyclo_len"], ""]]
            self.add_cyclo(node)
            del self.tmp["cyclo_yl"]
            del self.tmp["cyclo_len"]
            
        
        if "post_adding" in self.tmp:
            def add_position(func_group, pos):
                func_group.position += func_group.position >= pos
                if type(func_group) == Cycle:
                    func_group.start += func_group.start >= pos
                    func_group.end += func_group.end >= pos
                for fg_name, fg_list in func_group.functional_groups.items():
                    for fg in fg_list:
                        add_position(fg, pos)
            curr_fa = self.fatty_acyl_stack[-1]
            curr_fa.num_carbon += len(self.tmp["post_adding"])
            for pos in self.tmp["post_adding"]:
                add_position(curr_fa, pos)
                if type(curr_fa.double_bonds) == dict:
                    curr_fa.double_bonds = {(k + (k >= pos)): v for k, v in curr_fa.double_bonds.items()}
        
        if type(self.fatty_acyl_stack[-1].double_bonds) == dict and len(self.fatty_acyl_stack[-1].double_bonds) > 0:
            if sum(len(ct) > 0 for p, ct in self.fatty_acyl_stack[-1].double_bonds.items()) != len(self.fatty_acyl_stack[-1].double_bonds):
                self.set_lipid_level(LipidLevel.STRUCTURAL_SUBSPECIES)
                
        
        
        
    def set_acetyl(self, node):
        self.fatty_acyl_stack[-1] = FattyAcid("FA", num_carbon = 2)
        
        
        
    def build_GL(self, node):
        self.headgroup = self.tmp["gl_class"]
        del self.tmp["gl_class"]
        del self.tmp["generic_pos"]
        if len(self.fatty_acyl_dict) == 1: self.headgroup += "MG"
        if len(self.fatty_acyl_dict) == 2: self.headgroup += "DG"
        if len(self.fatty_acyl_dict) == 3: self.headgroup += "TG"
        self.build_GL_GP_SP(None)

        
        
        
    def build_GL_GP_SP(self, node):
        lipid_class = get_class(self.headgroup)
        if lipid_class == UNDEFINED_LIPID_CLASS:
            raise LipidParsingException("Unknown lipid class '%s'." % self.headgroup)
        
        self.fatty_acyl_list = [FattyAcid("FA%i" % i) for i in range(1, all_lipids[lipid_class]["max_fa"] + 1)]

        for fa_pos, fa in self.fatty_acyl_dict.items():
            self.fatty_acyl_list[fa_pos - 1] = fa
        
        
        
    def set_generic_fa(self, node):
        self.tmp["generic_pos"] = []
        
        
        
    def set_generic_pos(self, node):
        self.tmp["generic_pos"].append(int(node.get_text()))
        
       
       
    def add_generic_fa(self, node):
        if len(self.fatty_acyl_stack) != 1:
            raise LipidParsingException("Unable to parse lipid.")
                
        fa = self.fatty_acyl_stack.pop()
        
        fa_type = self.tmp["fa1"]["fa_type"]
        if (len(fa_type) == 2 and fa_type[-2:] == "yl") or (len(fa_type) >= 3 and fa_type[-3:] != "oyl"):
            fa.lipid_FA_bond_type = LipidFaBondType.ETHER_PLASMANYL
            if type(fa.double_bonds) == dict and 1 in fa.double_bonds:
                del fa.double_bonds[1]
                fa.lipid_FA_bond_type = LipidFaBondType.ETHER_PLASMENYL
                
        for pos in self.tmp["generic_pos"]:
            fa_copy = fa.copy()
            fa_copy.name = "FA%i" % pos
            self.fatty_acyl_dict[pos] = fa_copy
        self.tmp["generic_pos"] = []
        
        
        
    def set_car(self, node):
        self.tmp["fg_pos"] = []
        self.tmp["fg_type"] = ""
        
       
       
    def set_furan(self, node):
        self.tmp["furan"] = True
        
        
        
    def add_ol(self, node):
        self.tmp["fa1"]["fa_type"] = "FOH"
        
        
        
    def homo(self, node):
        self.tmp["post_adding"] = list(p[0] for p in self.tmp["fg_pos"])
        
        
        
    def add_summary(self, node):
        fa_i = "fa%i" % len(self.fatty_acyl_stack)
        self.tmp[fa_i]["fg_pos_summary"] = {k: v.upper() for k, v in self.tmp["fg_pos"]}
        
        
        
    def set_acetic_acid(self, node):
        self.fatty_acyl_stack[-1].num_carbon += 2
        self.tmp["fa1"]["fa_type"] = "FA"
        
        
        
    def add_func_stereo(self, node):
        self.tmp["fg_pos"][-1][1] = node.get_text()
        
        
        
    def set_yl_ending(self, node):
        l = int(node.get_text()) - 1
        if l == 0: return
    
        curr_fa = self.fatty_acyl_stack[-1]
        if l == 1:
            fname = "Me"
            fg = get_functional_group(fname)
        elif l == 2:
            fname = "Et"
            fg = get_functional_group(fname)
        else:
            fa = FattyAcid("FA", num_carbon = l)
            # shift functional groups
            for fg_name, fg_list in curr_fa.functional_groups.items():
                remove_item = []
                for i, func_group in enumerate(fg_list):
                    if func_group.position <= l:
                        remove_item.append(i)
                        if fg_name not in fa.functional_groups: fa.functional_groups[fg_name] = []
                        func_group.position = l + 1 - func_group.position
                        fa.functional_groups[fg_name].append(func_group)
                for i in remove_item[::-1]:
                    del curr_fa.functional_groups[fg_name][i]
            curr_fa.functional_groups = {k: v for k, v in curr_fa.functional_groups.items() if len(v) > 0}
            
            #shift double bonds
            if type(curr_fa.double_bonds) == dict:
                fa.double_bonds = {l + 1 - k: v for k, v in curr_fa.double_bonds.items() if k <= l}
                for k in list(k for k in curr_fa.double_bonds if k <= l): del curr_fa.double_bonds[k]
            
            fname = "cc"
            fg = CarbonChain(fa)
            
        curr_fa.num_carbon -= l
        fg.position = l
        curr_fa.shift_positions(-l)
        if fname not in curr_fa.functional_groups: curr_fa.functional_groups[fname] = []
        curr_fa.functional_groups[fname].append(fg)
        
        
        
    def set_methylene(self, node):
        self.tmp["fg_type"] = "methylene"
        if len(self.tmp["fg_pos"]) > 1:
            if self.tmp["fg_pos"][0][0] < self.tmp["fg_pos"][1][0]: self.tmp["fg_pos"][1][0] += 1
            elif self.tmp["fg_pos"][0][0] > self.tmp["fg_pos"][1][0]: self.tmp["fg_pos"][0][0] += 1
            self.fatty_acyl_stack[-1].num_carbon += 1
            self.tmp["add_methylene"] = True
     
    
    def check_db(self, node):
        fa_i = "fa%i" % len(self.fatty_acyl_stack)
        curr_fa = self.fatty_acyl_stack[-1]
        if "fg_pos_summary" in self.tmp[fa_i]:
            if type(curr_fa.double_bonds) != dict: curr_fa.double_bonds = {}
            for k, v in self.tmp[fa_i]["fg_pos_summary"].items():
                if v in {"E", "Z", ""} and k > 0 and k not in curr_fa.double_bonds: curr_fa.double_bonds[k] = v
                    
        
        
    def add_car(self, node):
        self.tmp["fa1"]["fa_type"] = "CAR"
        
        
        
    def add_hydroxyl(self, node):
        self.tmp["hydroxyl_pos"].append(int(node.get_text()))
    
    
    
    def set_coa(self, node):
        self.tmp["fa1"]["fa_type"] = "CoA"
        
        
        
    def set_methyl(self, node):
        self.fatty_acyl_stack[-1].num_carbon += 1
    
    
    
    def setup_hydroxyl(self, node):
        self.tmp["hydroxyl_pos"] = []
        
        
        
    def set_iso(self, node):
        curr_fa = self.fatty_acyl_stack[-1]
        curr_fa.num_carbon -= 1
        fg = get_functional_group("Me")
        fg.position = 2
        if "Me" not in curr_fa.functional_groups: curr_fa.functional_groups["Me"] = []
        curr_fa.functional_groups["Me"].append(fg)
        
        
        
    def set_ate(self, node):
        self.fatty_acyl_stack[-1].num_carbon += ate[node.get_text()]
        self.tmp["fa1"]["fa_type"] = "WE"
        
        
        
    def add_amine(self, node):
        fa_i = "fa%i" % len(self.fatty_acyl_stack)
        fa = self.fatty_acyl_stack.pop()
        
        fa.name += "1"
        self.fatty_acyl_stack[-1].name += "2"
        fa.lipid_FA_bond_type = LipidFaBondType.AMINE
        self.fatty_acyl_stack.insert(0, fa)
        
        
        
    def add_amine_name(self, node):
        self.tmp["fa1"]["fa_type"] = "NA"
        
        
        
    def add_ethanolamine(self, node):
        self.tmp["fa1"]["fa_type"] = "NAE"
        
        
    
    def add_hydroxyls(self, node):
        if len(self.tmp["hydroxyl_pos"]) > 1:
            fg_oh = get_functional_group("OH")
            for pos in sorted(self.tmp["hydroxyl_pos"], reverse = True)[:-1]:
                fg_insert = fg_oh.copy()
                fg_insert.position = pos
                if "OH" not in self.fatty_acyl_stack[-1].functional_groups: self.fatty_acyl_stack[-1].functional_groups["OH"] = []
                self.fatty_acyl_stack[-1].functional_groups["OH"].append(fg_insert)
        
        
        
    def set_double_bond_position(self, node):
        fa_i = "fa%i" % len(self.fatty_acyl_stack)
        pos = int(node.get_text())
        self.tmp[fa_i]["db_position"] = pos - (sum([1 for p in self.tmp["reduction"] if p < pos]) if "reduction" in self.tmp else 0)
        
        
        
    def set_recursion_pos(self, node):
        fa_i = "fa%i" % len(self.fatty_acyl_stack)
        self.tmp[fa_i]["recursion_pos"] = int(node.get_text())
        
        
        
    def set_recursion(self, node):
        self.tmp["fg_pos"] = []
        self.tmp["fg_type"] = ""
        self.fatty_acyl_stack.append(FattyAcid("FA"))
        fa_i = "fa%i" % len(self.fatty_acyl_stack)
        self.tmp[fa_i] = {"recursion_pos": 0}
        
        
        
    def add_recursion(self, node):
        fa_i = "fa%i" % len(self.fatty_acyl_stack)
        pos = self.tmp[fa_i]["recursion_pos"]
        
        
        fa = self.fatty_acyl_stack.pop()
        fa.position = pos
        curr_fa = self.fatty_acyl_stack[-1]
        
        if "cyclo_yl" in self.tmp:
            fname = "cyclo"
            del self.tmp["cyclo_yl"]
        else:
            fname = self.tmp["fa1"]["fa_type"]
        if fname not in curr_fa.functional_groups: curr_fa.functional_groups[fname] = []
        curr_fa.functional_groups[fname].append(fa)
        
        self.tmp["added_func_group"] = True
        
        
        
    def add_wax_ester(self, node):
        fa_i = "fa%i" % len(self.fatty_acyl_stack)
        fa = self.fatty_acyl_stack.pop()
        
        fa.name += "1"
        fa.lipid_FA_bond_type = LipidFaBondType.AMINE
        self.fatty_acyl_stack[-1].name += "2"
        self.fatty_acyl_stack.insert(0, fa)
        
        
        
    def set_cycle(self, node):
        self.tmp["cyclo"] = True
        
        
        
    def set_fatty_acid(self, node):
        def switch_position(func_group, switch):
            func_group.position = switch - func_group.position
            for fg_name, fg_list in func_group.functional_groups.items():
                for fg in fg_list:
                    switch_position(fg, switch)
            
        curr_fa = self.fatty_acyl_stack[-1]
        
        if "noyloxy" in curr_fa.functional_groups:
            if self.tmp["fa1"]["fa_type"] == "FA": self.tmp["fa1"]["fa_type"] = "FAHFA"
            
            while len(curr_fa.functional_groups["noyloxy"]) > 0:
                fa = curr_fa.functional_groups["noyloxy"].pop()
            
                
                acyl = AcylAlkylGroup(fa)
                acyl.position = fa.position
                
                if "acyl" not in curr_fa.functional_groups: curr_fa.functional_groups["acyl"] = []
                curr_fa.functional_groups["acyl"].append(acyl)
            del curr_fa.functional_groups["noyloxy"]
            
            
        elif "nyloxy" in curr_fa.functional_groups:
            while len(curr_fa.functional_groups["nyloxy"]) > 0:
                fa = curr_fa.functional_groups["nyloxy"].pop()
            
                
                alkyl = AcylAlkylGroup(fa, alkyl = True)
                alkyl.position = fa.position
                
                if "alkyl" not in curr_fa.functional_groups: curr_fa.functional_groups["alkyl"] = []
                curr_fa.functional_groups["alkyl"].append(alkyl)
            del curr_fa.functional_groups["nyloxy"]
            
                
        elif sum([k[-2:] == "yl" for k in curr_fa.functional_groups]) > 0:
            while True:
                try:
                    yl = [k for k in curr_fa.functional_groups if k[-2:] == "yl"][0]
                except Exception:
                    break
            
                while len(curr_fa.functional_groups[yl]) > 0:
                    fa = curr_fa.functional_groups[yl].pop()
                    
                    if "furan" in self.tmp:
                        self.tmp['cyclo_len'] = 4
                        fa.shift_positions(4)
                        for fg, fg_list in fa.functional_groups.items():
                            if fg not in curr_fa.functional_groups: curr_fa.functional_groups[fg] = fg_list
                            else: curr_fa.functional_groups[fg] += fg_list
                            
                        curr_fa.num_carbon += 4 + fa.num_carbon
                        
                        if type(curr_fa.double_bonds) == int: curr_fa.double_bonds = {}
                        if type(fa.double_bonds) == dict:
                            for pos, ez in fa.double_bonds.items():
                                curr_fa.double_bonds[pos + cyclo_len] = ez
                        curr_fa.double_bonds[1] = "E"
                        curr_fa.double_bonds[3] = "E"
                        self.tmp["cyclo_yl"] = True
                        
                        
                    
                    elif "cyclo" in self.tmp:
                        cyclo_len = curr_fa.num_carbon
                        self.tmp['cyclo_len'] = cyclo_len
                        switch_position(curr_fa, 2 + cyclo_len)
                        if type(curr_fa.double_bonds) == dict: curr_fa.double_bonds = {2 + cyclo_len - k: v for k, v in curr_fa.double_bonds.items()}
                        fa.shift_positions(cyclo_len)
                        
                        for fg, fg_list in fa.functional_groups.items():
                            if fg not in curr_fa.functional_groups: curr_fa.functional_groups[fg] = fg_list
                            else: curr_fa.functional_groups[fg] += fg_list
                            
                        
                        curr_fa.num_carbon = cyclo_len + fa.num_carbon
                        
                        if type(curr_fa.double_bonds) == int: curr_fa.double_bonds = {}
                        if type(fa.double_bonds) == dict:
                            for pos, ez in fa.double_bonds.items():
                                curr_fa.double_bonds[pos + cyclo_len] = ez
                        self.tmp["cyclo_yl"] = True
                        
                    else:
                        ## add carbon chains here here
                        ## special chains: i.e. ethyl, methyl
                        fg_name = ""
                        if (fa.double_bonds if type(fa.double_bonds) == int else len(fa.double_bonds)) == 0 and len(fa.functional_groups) == 0:
                            if fa.num_carbon == 1:
                                fg_name = "Me"
                                fg = get_functional_group(fg_name)
                            
                            elif fa.num_carbon == 2:
                                fg_name = "Et"
                                fg = get_functional_group(fg_name)
                                
                            if len(fg_name) > 0:
                                fg.position = fa.position
                                if fg_name not in curr_fa.functional_groups: curr_fa.functional_groups[fg_name] = []
                                curr_fa.functional_groups[fg_name].append(fg)
                            
                        if len(fg_name) == 0:
                            cc = CarbonChain(fa, position = fa.position)
                            if "cc" not in curr_fa.functional_groups: curr_fa.functional_groups["cc"] = []
                            curr_fa.functional_groups["cc"].append(cc)
                            
                if "cyclo" in self.tmp: del self.tmp["cyclo"]
                del curr_fa.functional_groups[yl]
           
           
        if "cyclo" in curr_fa.functional_groups:
            fa = curr_fa.functional_groups["cyclo"][0]
            del curr_fa.functional_groups["cyclo"]
            start_pos, end_pos = curr_fa.num_carbon + 1, curr_fa.num_carbon + (self.tmp["cyclo_len"] if 'cyclo_len' in self.tmp else 5)
            fa.shift_positions(start_pos - 1)
            
            if "cy" in curr_fa.functional_groups:
                for cy in curr_fa.functional_groups["cy"]:
                    cy.shift_positions(start_pos - 1)
            
            for fg, fg_list in fa.functional_groups.items():
                if fg not in curr_fa.functional_groups: curr_fa.functional_groups[fg] = fg_list
                else: curr_fa.functional_groups[fg] += fg_list
            
            curr_fa.num_carbon += fa.num_carbon
            
            if type(curr_fa.double_bonds) == int: curr_fa.double_bonds = {}
            if type(fa.double_bonds) == dict:
                for pos, ez in fa.double_bonds.items():
                    curr_fa.double_bonds[pos + start_pos - 1] = ez
                    
            self.tmp["fg_pos"] = [[start_pos, ""], [end_pos, ""]]
            self.add_cyclo(node)
            
            if "cyclo_len" in self.tmp: del self.tmp["cyclo_len"]
            if "cyclo" in self.tmp: del self.tmp["cyclo"]
            if "furan" in self.tmp: del self.tmp["furan"]
            
            
        elif "cyclo" in self.tmp:
            self.tmp["fg_pos"] = [[1, ""], [curr_fa.num_carbon, ""]]
            self.add_cyclo(node)
            del self.tmp["cyclo"]
        
            
        
    def set_double_bond_information(self, node):
        fa_i = "fa%i" % len(self.fatty_acyl_stack)
        self.tmp[fa_i]["db_position"] = 0
        self.tmp[fa_i]["db_cistrans"] = ""
        
        
        
    def reduction(self, node):
        self.fatty_acyl_stack[-1].num_carbon -= len(self.tmp["fg_pos"])
        self.tmp["reduction"] = [p[0] for p in self.tmp["fg_pos"]]
        
        
        
    def add_double_bond_information(self, node):
        fa_i = "fa%i" % len(self.fatty_acyl_stack)
        pos = self.tmp[fa_i]["db_position"]
        cistrans = self.tmp[fa_i]["db_cistrans"].upper()
        if cistrans == "" and "fg_pos_summary" in self.tmp[fa_i] and pos in self.tmp[fa_i]["fg_pos_summary"]: cistrans = self.tmp[fa_i]["fg_pos_summary"][pos]
        if pos == 0: return
        
        del self.tmp[fa_i]["db_position"]
        del self.tmp[fa_i]["db_cistrans"]
        if type(self.fatty_acyl_stack[-1].double_bonds) == int: self.fatty_acyl_stack[-1].double_bonds = {}
        
        if pos not in self.fatty_acyl_stack[-1].double_bonds or len(self.fatty_acyl_stack[-1].double_bonds[pos]) == 0:
            self.fatty_acyl_stack[-1].double_bonds[pos] = cistrans
        
        
        
    def set_dioic(self, node):
        self.tmp["fa1"]["fa_type"] = "FA"
        
        pos = self.tmp["fg_pos"][1][0] if len(self.tmp["fg_pos"]) == 2 else self.fatty_acyl_stack[-1].num_carbon
        self.fatty_acyl_stack[-1].num_carbon -= 1
        func_group = get_functional_group("COOH")
        func_group.position = pos - 1
        if "COOH" not in self.fatty_acyl_stack[-1].functional_groups: self.fatty_acyl_stack[-1].functional_groups["COOH"] = []
        self.fatty_acyl_stack[-1].functional_groups["COOH"].append(func_group)
        
        
        
    def set_cistrans(self, node):
        self.tmp["fa%i" % len(self.fatty_acyl_stack)]["db_cistrans"] = node.get_text()
        
        
        
    def set_fatty_acyl_type(self, node):
        t = node.get_text()
        
        if t[-2:] == "ol": self.tmp["fa1"]["fa_type"] = "FOH"
        elif t in {"noic acid", "nic acid", "dioic_acid"}: self.tmp["fa1"]["fa_type"] = "FA"
        elif t in {"nal", "dial"}: self.tmp["fa1"]["fa_type"] = "FAL"
        elif t in {"acetate", "noate", "nate"}: self.tmp["fa1"]["fa_type"] = "WE"
        elif t == "ne":
            self.tmp["fa1"]["fa_type"] = "HC"
            self.fatty_acyl_stack[-1].lipid_FA_bond_type = LipidFaBondType.AMINE
            
        else: self.tmp["fa1"]["fa_type"] = t
        
        
        
    def set_lipid_level(self, level):
        self.level = self.level if self.level.value < level.value else level
        
        
        
    def set_dial(self, node):
        curr_fa = self.fatty_acyl_stack[-1]
        pos = curr_fa.num_carbon
        fg = get_functional_group("oxo")
        fg.position = pos
        if "oxo" not in curr_fa.functional_groups: curr_fa.functional_groups["oxo"] = []
        curr_fa.functional_groups["oxo"].append(fg)
        
        
    
    def reset_length(self, node):
        self.tmp["length"] = 0
    
    
    
    def set_functional_length(self, node):
        if self.tmp["length"] != len(self.tmp["fg_pos"]):
            raise LipidException("Length of functional group '%i' does not match with number of its positions '%i'" % (self.tmp["length"], len(self.tmp["fg_pos"])))
    
    
    
    def set_fatty_length(self, node):
        self.fatty_acyl_stack[-1].num_carbon += self.tmp["length"]
        
        
        
    def rearrange_cycle(self, node):
        if "post_adding" in self.tmp:
            self.fatty_acyl_stack[-1].num_carbon += len(self.tmp["post_adding"])
            del self.tmp["post_adding"]
            
        curr_fa = self.fatty_acyl_stack[-1]
        start = self.tmp["fg_pos"][0][0]
        if "cy" in curr_fa.functional_groups:
            for cy in curr_fa.functional_groups["cy"]:
                shift = start - cy.position
                if shift == 0: continue
                cy.rearrange_functional_groups(curr_fa, shift)
        
        
        
    def add_cyclo(self, node):
        start = self.tmp["fg_pos"][0][0]
        end = self.tmp["fg_pos"][1][0]
        
            
        cyclo_db = None
        # check double bonds
        if type(self.fatty_acyl_stack[-1].double_bonds) == dict and len(self.fatty_acyl_stack[-1].double_bonds) > 0:
            cyclo_db = {db_pos: val for db_pos, val in self.fatty_acyl_stack[-1].double_bonds.items() if start <= db_pos <= end}
            
            for pos in cyclo_db:
                del self.fatty_acyl_stack[-1].double_bonds[pos]
                
        # check functional_groups
        cyclo_fg, remove_list, curr_fa = {}, set(), self.fatty_acyl_stack[-1]
        
        if "noyloxy" in curr_fa.functional_groups:
            remove_item = []
            for i, func_group in enumerate(curr_fa.functional_groups["noyloxy"]):
                if start <= func_group.position <= end:
                    cc = CarbonChain(func_group, position = func_group.position)
                    
                    if "cc" not in curr_fa.functional_groups: curr_fa.functional_groups["cc"] = []
                    curr_fa.functional_groups["cc"].append(cc)
                    remove_item.append(i)
                    
            for i in remove_item[::-1]: del curr_fa.functional_groups["noyloxy"][i]
            if len(curr_fa.functional_groups["noyloxy"]) == 0: remove_list.add("noyloxy")
            
        for fg, fg_list in curr_fa.functional_groups.items():
            remove_item = []
            
            for i, func_group in enumerate(fg_list):
                if start <= func_group.position <= end:
                    if fg not in cyclo_fg: cyclo_fg[fg] = []
                    cyclo_fg[fg].append(func_group)
                    remove_item.append(i)
                    
            for i in remove_item[::-1]: del curr_fa.functional_groups[fg][i]
            if len(fg_list) == 0: remove_list.add(fg)
            
        for fg in remove_list: del curr_fa.functional_groups[fg]
        cy = end - start + 1
        
        bridge_chain = []
        if "furan" in self.tmp:
            bridge_chain.append(Element.O)
            cy += 1

        cycle = Cycle(cy, start = start, end = end, double_bonds = cyclo_db, functional_groups = cyclo_fg, bridge_chain = bridge_chain)
        if "cy" not in self.fatty_acyl_stack[-1].functional_groups: self.fatty_acyl_stack[-1].functional_groups["cy"] = []
        self.fatty_acyl_stack[-1].functional_groups["cy"].append(cycle)
        
        
        
    def add_epoxy(self, node):
        self.tmp["fg_pos"] = self.tmp["fg_pos"][:1]
        self.tmp["fg_type"] = "Epoxy"
        
    
    
    def special_number(self, node):
        self.tmp["length"] += special_numbers[node.get_text()]
        
        
        
    def last_number(self, node):
        self.tmp["length"] += last_numbers[node.get_text()]
        
        
        
    def second_number(self, node):
        self.tmp["length"] += second_numbers[node.get_text()]
        
        
        
    def set_functional_group(self, node):
        self.tmp["fg_pos"] = []
        self.tmp["fg_type"] = ""
        
        
        
    def set_prosta(self, node):
        minus_pos = (sum([1 for p in self.tmp["reduction"] if p < 8]) if "reduction" in self.tmp else 0)
        self.tmp["fg_pos"] = [[8 - minus_pos, ""], [12 - minus_pos, ""]]
        self.tmp["fg_type"] = "cy"
        
        
        
    def add_functional_group(self, node):
        if "added_func_group" in self.tmp: 
            del self.tmp["added_func_group"]
            return
        
        elif "add_methylene" in self.tmp: 
            del self.tmp["add_methylene"]
            self.add_cyclo(node)
            return
        
        t = self.tmp["fg_type"]
        
        if t != "acetoxy":
            if t not in func_groups: raise LipidException("Unknown functional group: '%s'" % t)
            t = func_groups[t]
            if len(t) == 0: return
            fg = get_functional_group(t)
        else:
            fg = AcylAlkylGroup(FattyAcid("O", num_carbon = 2))
            
            
        
        if t not in self.fatty_acyl_stack[-1].functional_groups: self.fatty_acyl_stack[-1].functional_groups[t] = []
        for pos in self.tmp["fg_pos"]:
            fg_insert = fg.copy()
            fg_insert.position = pos[0] - (sum([1 for p in self.tmp["reduction"] if p < pos[0]]) if "reduction" in self.tmp else 0)
            self.fatty_acyl_stack[-1].functional_groups[t].append(fg_insert)
        
        
        
    def set_functional_position(self, node):
        self.tmp["fg_pos"].append([0, ""])
        
        
        
    def set_functional_pos(self, node):
        self.tmp["fg_pos"][-1][0] = int(node.get_text())
        
        
        
    def set_functional_type(self, node):
        self.tmp["fg_type"] = node.get_text()
        
        
        
    def build_lipid(self, node):
        lipid_level_class = None
        if self.level == LipidLevel.ISOMERIC_SUBSPECIES: lipid_level_class = LipidIsomericSubspecies
        if self.level == LipidLevel.STRUCTURAL_SUBSPECIES: lipid_level_class = LipidStructuralSubspecies
        if self.level == LipidLevel.MOLECULAR_SUBSPECIES: lipid_level_class = LipidMolecularSubspecies
        if self.level == LipidLevel.SPECIES: lipid_level_class = LipidSpecies
        
        headgroup = HeadGroup(self.headgroup)
        
        self.content = LipidAdduct()
        self.content.lipid = lipid_level_class(headgroup, self.fatty_acyl_list)
            



"""
MIT License

Copyright (c) 2020 Dominik Kopczynski   -   dominik.kopczynski {at} univie.ac.at
                   Nils Hoffmann  -  nils.hoffmann {at} cebitec.uni-bielefeld.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


from pygoslin.domain.LipidExceptions import RuntimeException
from pygoslin.domain.LipidLevel import LipidLevel
from pygoslin.domain.LipidCategory import LipidCategory
from pygoslin.domain.LipidExceptions import *
from pygoslin.domain.LipidSpeciesInfo import LipidSpeciesInfo
from pygoslin.domain.LipidClass import *
from pygoslin.domain.FunctionalGroup import *
from pygoslin.domain.LipidFaBondType import *
from pygoslin.domain.Element import Element

glyco_table = {"ga1": ["Gal", "GalNAc", "Gal", "Glc"],
               "ga2": ["GalNAc", "Gal", "Glc"],
               "gb3": ["Gal", "Gal", "Glc"],
               "gb4": ["GalNAc", "Gal", "Gal", "Glc"],
               "gd1": ["Gal", "GalNAc", "NeuAc", "NeuAc", "Gal", "Glc"],
               "gd1a": ["Hex", "Hex", "Hex", "HexNAc", "NeuAc", "NeuAc"],
               "gd2": ["GalNAc", "NeuAc", "NeuAc", "Gal", "Glc"],
               "gd3": ["NeuAc", "NeuAc", "Gal", "Glc"],
               "gm1": ["Gal", "GalNAc", "NeuAc", "Gal", "Glc"],
               "gm2": ["GalNAc", "NeuAc", "Gal", "Glc"],
               "gm3": ["NeuAc", "Gal", "Glc"],
               "gm4": ["NeuAc", "Gal"],
               "gp1": ["NeuAc", "NeuAc", "Gal", "GalNAc", "NeuAc", "NeuAc", "NeuAc", "Gal", "Glc"],
               "gq1": ["NeuAc", "Gal", "GalNAc", "NeuAc", "NeuAc", "NeuAc", "Gal", "Glc"],
               "gt1": ["Gal", "GalNAc", "NeuAc", "NeuAc", "NeuAc", "Gal", "Glc"],
               "gt2": ["GalNAc", "NeuAc", "NeuAc", "NeuAc", "Gal", "Glc"],
               "gt3": ["NeuAc", "NeuAc", "NeuAc", "Gal", "Glc"],
               
               "gd0a": ["HexNAc", "Hex", "NeuAc", "HexNAc", "Hex", "NeuAc", "Hex"],
               "gd1b": ["Gal", "GalNAc", "NeuAc", "NeuAc", "Gal", "Glc"],
               "gq1b": ["NeuAc", "NeuAc", "Gal", "GalNAc", "NeuAc", "NeuAc", "Gal", "Glc"],
               "gt1b": ["NeuAc", "Gal", "GalNAc", "NeuAc", "NeuAc", "Gal", "Glc"],
               "gd1a-ac": ["Hex", "Hex", "Hex", "HexNAc", "NeuAc", "NeuAc", "NeuAc"],
               "gq1-ac": ["NeuAc", "Gal", "GalNAc", "NeuAc", "NeuAc", "NeuAc", "NeuAc", "Gal", "Glc"],
               "gt1b-ac": ["NeuAc", "Gal", "GalNAc", "NeuAc", "NeuAc", "NeuAc", "Gal", "Glc"],
               "gt3-ac": ["NeuAc", "NeuAc", "NeuAc", "NeuAc", "NeuAc", "Gal", "Glc"],
               
               }

class HeadGroup:
    def __init__(self, headgroup, decorators = None, use_headgroup = False):
        self.imported_headgroup = headgroup
        self.decorators = [d for d in decorators] if decorators != None else []
        self.parsed_headgroup = get_class(headgroup)
        if self.parsed_headgroup == UNDEFINED_LIPID_CLASS:
            self.parsed_headgroup = headgroup
        else:
            self.parsed_headgroup = all_lipids[self.parsed_headgroup]["name"]
        
        # checking if head group is a glyco-sphingolipid
        hg = headgroup.strip(" ").lower()
        if hg in glyco_table and not use_headgroup:
            for carbohydrate in glyco_table[hg]:
                try:
                    
                    functional_group = get_functional_group(carbohydrate)
                    functional_group.elements[Element.O] -= 1
                    self.decorators.append(functional_group)
                except Exception:
                    raise LipidParsingException("Carbohydrate '%s' unknown" % carbohydrate)
            headgroup = "Cer"
        
        self.headgroup = headgroup.strip(" ")
        self.lipid_category = get_category(self.headgroup)
        self.lipid_class = get_class(self.headgroup)
        self.use_headgroup = use_headgroup
        self.sp_exception = self.lipid_category == LipidCategory.SP and all_lipids[self.lipid_class]["name"] in {"Cer", "SPB"} and len(self.decorators) == 0
        
        
    def get_lipid_string(self, level = None):
        if level == LipidLevel.CATEGORY:
            return self.lipid_category.name
        
        headgoup_string = [all_lipids[self.lipid_class]["name"] if not self.use_headgroup else self.headgroup]
        prefix = []
                
        if len(self.decorators) > 0:
            if level not in {LipidLevel.COMPLETE_STRUCTURE, LipidLevel.FULL_STRUCTURE, LipidLevel.STRUCTURE_DEFINED}:
                decorators_sorted = []
                for hgd in self.decorators:
                    if hgd.suffix: continue
                    
                    hgd_copy = hgd.copy()
                    if hgd_copy.name.find("Gal") > -1 or hgd_copy.name.find("Glc") > -1 or hgd_copy.name.find("S(3')") > -1:
                        hgd_copy.name = hgd_copy.name.replace("Gal", "Hex").replace("Glc", "Hex").replace("S(3')", "S")
                    decorators_sorted.append(hgd_copy)
                
                if len(decorators_sorted) > 0:
                    decorators_sorted.sort(key = lambda x: x.name)
                    name, rep = "", 0
                    
                    for hgd in decorators_sorted:
                        if name != hgd.name:
                            if rep > 1: prefix.append("%i" % rep)
                            name, rep = hgd.name, hgd.count
                            prefix.append(name)
                        else:
                            rep += hgd.count
                        
                    if rep > 1: prefix.append("%i" % rep)
                
            else:
                prefix = ["%s-" % hgd.to_string(level) for hgd in self.decorators if not hgd.suffix]
            
        suffix = [hgd.to_string(level) for hgd in self.decorators if hgd.suffix]
        if level in {LipidLevel.COMPLETE_STRUCTURE, LipidLevel.FULL_STRUCTURE} and self.lipid_category == LipidCategory.SP and not self.sp_exception: suffix.append("(1)")
            
        return "".join(prefix + headgoup_string + suffix)
        
        
        
    def get_elements(self):
        if self.use_headgroup:
            raise LipidException("Element table cannot be computed for lipid '%s'" % self.headgroup)
        
        dummy = FunctionalGroup("dummy", elements = all_lipids[self.lipid_class]["elements"])
        for hgd in self.decorators: dummy += hgd
        
        return dummy.elements
